# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import json
import warnings

from momba import model
from momba.model import (
    functions,
    automata,
    context,
    distributions,
    expressions,
    operators,
    properties,
    types,
)

from .dump_model import ModelFeature


class JANIError(Exception):
    """
    Indicates an error while loading a JANI string.

    Usually either :class:`InvalidJANIError` or
    :class:`UnsupportedJANIError` gets thrown.
    """


class InvalidJANIError(JANIError):
    """
    String is not valid JANI.
    """


class UnsupportedJANIError(JANIError):
    """
    Model contains unsupported JANI features.

    Attributes
    ----------
    unsupported_features:
        Used model features not supported by Momba.
    """

    unsupported_features: t.AbstractSet[t.Union[str, ModelFeature]]

    def __init__(self, unsupported_features: t.AbstractSet[t.Union[str, ModelFeature]]):
        self.unsupported_features = unsupported_features
        super().__init__(f"unsupported JANI features {self.unsupported_features!r}")


_TYPE_MAP = {
    "bool": types.BOOL,
    "int": types.INT,
    "real": types.REAL,
    "clock": types.CLOCK,
    "continuous": types.CONTINUOUS,
}

_BINARY_OP_MAP: t.Mapping[str, expressions.BinaryConstructor] = {
    "∨": expressions.logic_or,
    "∧": expressions.logic_and,
    "⇒": expressions.logic_implies,  # defined by the `derived-operators` extension
    "⊕": expressions.logic_xor,  # defined by the `x-momba-operators` extension
    "⇔": expressions.logic_equiv,  # defined by the `x-momba-operators` extension
    "=": expressions.equals,
    "≠": expressions.not_equals,
    "<": expressions.less,
    "≤": expressions.less_or_equal,
    ">": expressions.greater,  # defined by the `derived-operators` extension
    "≥": expressions.greater_or_equal,  # defined by the `derived-operators` extension
    "+": expressions.add,
    "-": expressions.sub,
    "*": expressions.mul,
    "%": expressions.mod,
    "/": expressions.real_div,
    "//": expressions.floor_div,  # defined by the `x-momba-operators`extension
    "pow": expressions.power,
    "log": expressions.log,
    "min": expressions.minimum,  # defined by the `derived-operators` extension
    "max": expressions.maximum,  # defined by the `derived-operators` extension
}

_UNARY_OP_MAP: t.Mapping[str, expressions.UnaryConstructor] = {
    "¬": expressions.logic_not,
    "floor": expressions.floor,
    "ceil": expressions.ceil,
    "abs": expressions.absolute,  # defined by the `derived-operators` extension
    "trc": expressions.trunc,  # defined by the `derived-operators` extension
    "sgn": expressions.sgn,  # defined by the `derived-operators` extension
}


_AGGREGATION_FUNCTIONS: t.Mapping[str, operators.AggregationFunction] = {
    "min": operators.AggregationFunction.MIN,
    "max": operators.AggregationFunction.MAX,
    "sum": operators.AggregationFunction.SUM,
    "avg": operators.AggregationFunction.AVG,
    "count": operators.AggregationFunction.COUNT,
    "argmin": operators.AggregationFunction.ARGMIN,
    "argmax": operators.AggregationFunction.ARGMAX,
    "values": operators.AggregationFunction.VALUES,
    "∀": operators.AggregationFunction.FORALL,
    "∃": operators.AggregationFunction.EXISTS,
}

_ACCUMULATION_INSTANTS: t.Mapping[str, properties.AccumulationInstant] = {
    "steps": properties.AccumulationInstant.STEPS,
    "time": properties.AccumulationInstant.TIME,
    "exit": properties.AccumulationInstant.EXIT,
}


_STATE_SELECTORS: t.Mapping[str, properties.StateSelector] = {
    "initial": properties.INITIAL_STATES,
    "deadlock": properties.DEADLOCK_STATES,
    "timelock": properties.TIMELOCK_STATES,
}

_MIN_MAX_OPERATORS: t.Mapping[str, operators.MinMax] = {
    "Pmin": operators.MinMax.MIN,
    "Pmax": operators.MinMax.MAX,
    "Emin": operators.MinMax.MIN,
    "Emax": operators.MinMax.MAX,
    "Smin": operators.MinMax.MIN,
    "Smax": operators.MinMax.MAX,
}

_BINARY_PATH_OPERATORS: t.Mapping[str, operators.BinaryPathOperator] = {
    "U": operators.BinaryPathOperator.UNTIL,
    "W": operators.BinaryPathOperator.WEAK_UNTIL,
    "R": operators.BinaryPathOperator.RELEASE,
}

_UNARY_PATH_OPERATORS: t.Mapping[str, operators.UnaryPathOperator] = {
    "F": operators.UnaryPathOperator.EVENTUALLY,
    "G": operators.UnaryPathOperator.GLOBALLY,
}

_TRIGONOMETRIC_FUNCTIONS: t.Mapping[str, operators.TrigonometricFunction] = {
    function.value: function for function in operators.TrigonometricFunction
}


def _expression(jani_expression: t.Any) -> expressions.Expression:
    if isinstance(jani_expression, (float, bool, int)):
        if (
            isinstance(jani_expression, float)
            and int(jani_expression) == jani_expression
        ):
            jani_expression = int(jani_expression)
        return expressions.ensure_expr(jani_expression)
    elif isinstance(jani_expression, str):
        return expressions.name(jani_expression)
    elif isinstance(jani_expression, dict):
        if "constant" in jani_expression:
            return expressions.ensure_expr(jani_expression["constant"])
        elif "op" in jani_expression:
            op = jani_expression["op"]
            if op in _BINARY_OP_MAP:
                _check_fields(jani_expression, required={"op", "left", "right"})
                left = _expression(jani_expression["left"])
                right = _expression(jani_expression["right"])
                return _BINARY_OP_MAP[op](left, right)
            elif op in _UNARY_OP_MAP:
                _check_fields(jani_expression, required={"op", "exp"})
                operand = _expression(jani_expression["exp"])
                return _UNARY_OP_MAP[op](operand)
            elif op in _TRIGONOMETRIC_FUNCTIONS:
                _check_fields(jani_expression, required={"op", "exp"})
                operand = _expression(jani_expression["exp"])
                return expressions.Trigonometric(_TRIGONOMETRIC_FUNCTIONS[op], operand)
            elif op == "aa":
                _check_fields(jani_expression, required={"op", "exp", "index"})
                return expressions.ArrayAccess(
                    array=_expression(jani_expression["exp"]),
                    index=_expression(jani_expression["index"]),
                )
            elif op == "av":
                _check_fields(jani_expression, required={"op", "elements"})
                return expressions.ArrayValue(
                    tuple(map(_expression, jani_expression["elements"]))
                )
            elif op == "ac":
                _check_fields(jani_expression, required={"op", "var", "length", "exp"})
                return expressions.ArrayConstructor(
                    variable=jani_expression["var"],
                    length=_expression(jani_expression["length"]),
                    expression=_expression(jani_expression["exp"]),
                )
            elif op == "ite":
                _check_fields(jani_expression, required={"op", "if", "then", "else"})
                condition = _expression(jani_expression["if"])
                consequence = _expression(jani_expression["then"])
                alternative = _expression(jani_expression["else"])
                return expressions.ite(condition, consequence, alternative)
            elif op == "der":
                _check_fields(jani_expression, required={"op", "var"})
                variable = jani_expression["var"]
                return expressions.Derivative(variable)
            elif op == "call":
                _check_fields(jani_expression, required={"op", "function", "args"})
                return functions.CallExpression(
                    function=jani_expression["function"],
                    arguments=tuple(
                        _expression(argument) for argument in jani_expression["args"]
                    ),
                )
            elif op == "nondet":
                _check_fields(jani_expression, required={"op", "var", "exp"})
                return expressions.Selection(
                    variable=jani_expression["var"],
                    condition=_expression(jani_expression["exp"]),
                )
        elif "distribution" in jani_expression:
            _check_fields(jani_expression, required={"args", "distribution"})
            arguments = tuple(map(_expression, jani_expression["args"]))
            distribution = distributions.DistributionType.by_name(
                jani_expression["distribution"]
            )
            return expressions.Sample(distribution, arguments)
    return _property(jani_expression)


def _property_interval(jani_property: t.Any) -> properties.Interval:
    _check_fields(
        jani_property, optional={"lower", "lower-exclusive", "upper", "upper-exclusive"}
    )
    lower: t.Optional[expressions.Expression]
    lower_exclusive: t.Optional[expressions.Expression]
    upper: t.Optional[expressions.Expression]
    upper_exclusive: t.Optional[expressions.Expression]
    if "lower" in jani_property:
        lower = _expression(jani_property["lower"])
    else:
        lower = None
    if "lower-exclusive" in jani_property:
        lower_exclusive = _expression(jani_property["lower-exclusive"])
    else:
        lower_exclusive = None
    if "upper" in jani_property:
        upper = _expression(jani_property["upper"])
    else:
        upper = None
    if "upper-exclusive" in jani_property:
        upper_exclusive = _expression(jani_property["upper-exclusive"])
    else:
        upper_exclusive = None
    return properties.Interval(lower, lower_exclusive, upper, upper_exclusive)


def _reward_instant(jani_property: t.Any) -> properties.RewardInstant:
    _check_fields(jani_property, required={"exp", "accumulate", "instant"})
    return properties.RewardInstant(
        _expression(jani_property["exp"]),
        frozenset(_ACCUMULATION_INSTANTS[elem] for elem in jani_property["accumulate"]),
        _expression(jani_property["instant"]),
    )


def _reward_bound(jani_property: t.Any) -> properties.RewardBound:
    _check_fields(jani_property, required={"exp", "accumulate", "bounds"})
    return properties.RewardBound(
        _expression(jani_property["exp"]),
        frozenset(_ACCUMULATION_INSTANTS[elem] for elem in jani_property["accumulate"]),
        _property_interval(jani_property["bounds"]),
    )


def _property(jani_property: t.Any) -> expressions.Expression:
    if "op" not in jani_property:
        raise InvalidJANIError(f"{jani_property!r} is not a valid JANI property")
    if jani_property["op"] == "filter":
        _check_fields(jani_property, required={"op", "fun", "values", "states"})
        return properties.Aggregate(
            _AGGREGATION_FUNCTIONS[jani_property["fun"]],
            _expression(jani_property["values"]),
            _expression(jani_property["states"]),
        )
    if jani_property["op"] in {"Pmin", "Pmax"}:
        _check_fields(jani_property, required={"op", "exp"})
        return properties.Probability(
            _MIN_MAX_OPERATORS[jani_property["op"]], _expression(jani_property["exp"])
        )
    if jani_property["op"] in {"∀", "∃"}:
        _check_fields(jani_property, required={"op", "exp"})
        return properties.PathQuantifier(
            operators.Quantifier.FORALL
            if jani_property["op"] == "∀"
            else operators.Quantifier.EXISTS,
            _expression(jani_property["exp"]),
        )
    if jani_property["op"] in {"Emin", "Emax"}:
        _check_fields(
            jani_property,
            required={"op", "exp"},
            optional={
                "accumulate",
                "reach",
                "step-instant",
                "time-instant",
                "reward-instants",
            },
        )
        accumulate: t.Optional[t.FrozenSet[properties.AccumulationInstant]]
        reachability: t.Optional[expressions.Expression]
        step_instant: t.Optional[expressions.Expression]
        time_instant: t.Optional[expressions.Expression]
        reward_instants: t.Optional[t.Sequence[properties.RewardInstant]]
        if "accumulate" in jani_property:
            accumulate = frozenset(
                _ACCUMULATION_INSTANTS[elem] for elem in jani_property["accumulate"]
            )
        else:
            accumulate = None
        if "reach" in jani_property:
            reachability = _expression(jani_property["reach"])
        else:
            reachability = None
        if "step-instant" in jani_property:
            step_instant = _expression(jani_property["step-instant"])
        else:
            step_instant = None
        if "time-instant" in jani_property:
            time_instant = _expression(jani_property["time-instant"])
        else:
            time_instant = None
        if "reward-instants" in jani_property:
            reward_instants = [
                _reward_instant(elem) for elem in jani_property["reward-instants"]
            ]
        else:
            reward_instants = None
        return properties.ExpectedReward(
            operator=_MIN_MAX_OPERATORS[jani_property["op"]],
            reward=_expression(jani_property["exp"]),
            accumulate=accumulate,
            reachability=reachability,
            step_instant=step_instant,
            time_instant=time_instant,
            reward_instants=reward_instants,
        )
    if jani_property["op"] in {"Smin", "Smax"}:
        _check_fields(jani_property, required={"op", "exp"}, optional={"accumulate"})
        if "accumulate" in jani_property:
            return properties.SteadyState(
                _MIN_MAX_OPERATORS[jani_property["op"]],
                _expression(jani_property["exp"]),
                jani_property["accumulate"],
            )
        else:
            return properties.SteadyState(
                _MIN_MAX_OPERATORS[jani_property["op"]],
                _expression(jani_property["exp"]),
            )
    step_bounds: t.Optional[properties.Interval]
    time_bounds: t.Optional[properties.Interval]
    reward_bounds: t.Optional[t.Sequence[properties.RewardBound]]
    if jani_property["op"] in {"U", "W", "R"}:
        _check_fields(
            jani_property,
            required={"op", "left", "right"},
            optional={"step-bounds", "time-bounds", "reward-bounds"},
        )
        time_operator = _BINARY_PATH_OPERATORS[jani_property["op"]]
        if "step-bounds" in jani_property:
            step_bounds = _property_interval(jani_property["step-bounds"])
        else:
            step_bounds = None
        if "time-bounds" in jani_property:
            time_bounds = _property_interval(jani_property["time-bounds"])
        else:
            time_bounds = None
        if "reward-bounds" in jani_property:
            reward_bounds = list(map(_reward_bound, jani_property["reward-bounds"]))
        else:
            reward_bounds = None
        return properties.BinaryPathFormula(
            time_operator,
            _expression(jani_property["left"]),
            _expression(jani_property["right"]),
            step_bounds,
            time_bounds,
            reward_bounds,
        )
    if jani_property["op"] in _UNARY_PATH_OPERATORS:
        _check_fields(
            jani_property,
            required={"op", "exp"},
            optional={"step-bounds", "time-bounds", "reward-bounds"},
        )
        if "step-bounds" in jani_property:
            step_bounds = _property_interval(jani_property["step-bounds"])
        else:
            step_bounds = None
        if "time-bounds" in jani_property:
            time_bounds = _property_interval(jani_property["time-bounds"])
        else:
            time_bounds = None
        if "reward-bounds" in jani_property:
            reward_bounds = list(map(_reward_bound, jani_property["reward-bounds"]))
        else:
            reward_bounds = None
        return properties.UnaryPathFormula(
            _UNARY_PATH_OPERATORS[jani_property["op"]],
            _expression(jani_property["exp"]),
            step_bounds,
            time_bounds,
            reward_bounds,
        )
    if jani_property["op"] in _STATE_SELECTORS:
        return _STATE_SELECTORS[jani_property["op"]]
    raise ValueError(f"{jani_property} does not seem to be a valid JANI property")


def _type(typ: t.Any) -> types.Type:
    if isinstance(typ, str):
        return _TYPE_MAP[typ]
    assert isinstance(typ, dict)
    if typ["kind"] == "bounded":
        _check_fields(
            typ,
            required={"kind", "base"},
            optional={"lower-bound", "upper-bound"},
        )
        base = _type(typ["base"])
        assert isinstance(base, types.NumericType)
        lower_bound = _expression(typ["lower-bound"]) if "lower-bound" in typ else None
        upper_bound = _expression(typ["upper-bound"]) if "upper-bound" in typ else None
        return base.bound(lower_bound, upper_bound)
    elif typ["kind"] == "array":
        _check_fields(typ, required={"kind", "base"})
        return types.array_of(_type(typ["base"]))
    raise InvalidJANIError(f"{typ!r} is not a valid JANI type")


def _comment_warning(structure: t.Any) -> None:
    if "comment" in structure:
        warnings.warn(
            "comments are currently not supported, comment information will be lost"
        )


def _warn_fields(structure: t.Any, expected: t.Collection[str]) -> None:
    if hasattr(structure, "keys"):
        fields = set(structure.keys())
        for field in expected:
            fields.remove(field)
        if fields:
            for unknown in fields:
                warnings.warn(f"encountered unknown field {unknown} in {structure}")


_Fields = t.Collection[str]


def _check_fields(
    jani_structure: t.Any,
    required: _Fields = frozenset(),
    optional: _Fields = frozenset(),
    unsupported: _Fields = frozenset(),
) -> None:
    if not isinstance(jani_structure, dict):
        raise InvalidJANIError(f"expected map but found {jani_structure!r}")
    fields = {field for field in jani_structure.keys() if not field.startswith("x-")}
    for field in unsupported:
        if (field in optional or field in required) and field in fields:
            warnings.warn(
                f"field {field!r} in {jani_structure!r} is currently unsupported",
                stacklevel=2,
            )
    for field in required:
        if field not in fields:
            raise InvalidJANIError(
                f"field {field!r} is required but not found in {jani_structure!r}"
            )
        fields.discard(field)
    for field in optional:
        fields.discard(field)
    if fields:
        warnings.warn(f"unknown fields {fields!r} in {jani_structure!r}", stacklevel=2)


def _variable_declaration(jani_declaration: t.Any) -> context.VariableDeclaration:
    _check_fields(
        jani_declaration,
        required={"name", "type"},
        optional={"transient", "initial-value", "comment"},
    )
    transient: bool = jani_declaration.get("transient", None)
    initial_value: t.Optional[expressions.Expression] = (
        _expression(jani_declaration["initial-value"])
        if "initial-value" in jani_declaration
        else None
    )
    return context.VariableDeclaration(
        identifier=jani_declaration["name"],
        typ=_type(jani_declaration["type"]),
        is_transient=transient,
        initial_value=initial_value,
    )


def _constant_declaration(jani_declaration: t.Any) -> context.ConstantDeclaration:
    _check_fields(
        jani_declaration, required={"name", "type"}, optional={"value", "comment"}
    )
    value: t.Optional[expressions.Expression] = (
        _expression(jani_declaration["value"]) if "value" in jani_declaration else None
    )
    return context.ConstantDeclaration(
        identifier=jani_declaration["name"],
        typ=_type(jani_declaration["type"]),
        value=value,
    )


def _location(jani_location: t.Any) -> automata.Location:
    _check_fields(
        jani_location,
        required={"name"},
        optional={"time-progress", "transient-values", "x-momba-anonymous"},
    )
    progress_invariant: t.Optional[expressions.Expression]
    if "time-progress" in jani_location:
        _check_fields(
            jani_location["time-progress"], required={"exp"}, optional={"comment"}
        )
        progress_invariant = _expression(jani_location["time-progress"]["exp"])
    else:
        progress_invariant = None
    transient_values: t.Set[automata.Assignment] = set()
    if "transient-values" in jani_location:
        for jani_transient_value in jani_location["transient-values"]:
            _check_fields(
                jani_transient_value, required={"ref", "value"}, optional={"comment"}
            )
            assignment = automata.Assignment(
                target=_target(jani_transient_value["ref"]),
                value=_expression(jani_transient_value["value"]),
            )
            transient_values.add(assignment)
    return automata.Location(
        name=None
        if jani_location.get("x-momba-anonymous", False)
        else jani_location["name"],
        progress_invariant=progress_invariant,
        transient_values=frozenset(transient_values),
    )


_Locations = t.Dict[str, automata.Location]


def _target(jani_target: t.Any) -> expressions.Expression:
    if isinstance(jani_target, str):
        return expressions.Name(jani_target)
    _check_fields(jani_target, required={"op"}, optional={"exp", "index"})
    if jani_target["op"] == "aa":
        _check_fields(jani_target, required={"op", "exp", "index"})
        return expressions.ArrayAccess(
            array=_expression(jani_target["exp"]),
            index=_expression(jani_target["index"]),
        )
    raise InvalidJANIError(f"{jani_target!r} is not a valid lvalue")


def _edge(ctx: model.Context, locations: _Locations, jani_edge: t.Any) -> automata.Edge:
    _check_fields(
        jani_edge,
        required={"location", "destinations"},
        optional={"action", "rate", "guard", "comment"},
    )
    location = locations[jani_edge["location"]]
    action_pattern: t.Optional[model.ActionPattern] = None
    if "action" in jani_edge:
        if isinstance(jani_edge["action"], str):
            action_pattern = model.ActionPattern(
                ctx.get_action_type_by_name(jani_edge["action"])
            )
        else:
            _check_fields(
                jani_edge["action"], required={"name"}, optional={"arguments"}
            )
            action_pattern = model.ActionPattern(
                ctx.get_action_type_by_name(jani_edge["action"]["name"]),
                # FIXME: support import of action patterns
                # identifiers=jani_edge["action"].get("arguments", ()),
            )
    rate = _expression(jani_edge["rate"]["exp"]) if "rate" in jani_edge else None
    guard = _expression(jani_edge["guard"]["exp"]) if "guard" in jani_edge else None
    destinations = [
        automata.Destination(
            location=locations[jani_destination["location"]],
            probability=(
                _expression(jani_destination["probability"]["exp"])
                if "probability" in jani_destination
                else None
            ),
            assignments=tuple(
                automata.Assignment(
                    target=_target(jani_assignment["ref"]),
                    value=_expression(jani_assignment["value"]),
                    index=jani_assignment.get("index", 0),
                )
                for jani_assignment in jani_destination.get("assignments", ())
            ),
        )
        for jani_destination in jani_edge["destinations"]
    ]
    return automata.Edge(
        location=location,
        destinations=tuple(destinations),
        action_pattern=action_pattern,
        guard=guard,
        rate=rate,
    )


def _action_parameter(jani_action_parameter: t.Any) -> model.ActionParameter:
    _check_fields(
        jani_action_parameter,
        required={"type"},
        optional={"comment"},
    )
    typ = _type(jani_action_parameter["type"])
    comment = jani_action_parameter.get("comment", None)
    return model.ActionParameter(typ, comment=comment)


def _action(jani_action: t.Any) -> model.ActionType:
    _check_fields(jani_action, required={"name"}, optional={"comment", "parameters"})
    name = jani_action["name"]
    comment = jani_action.get("comment", None)
    parameters: t.List[model.ActionParameter] = []
    if "parameters" in jani_action:
        for jani_parameter in jani_action["parameters"]:
            parameters.append(_action_parameter(jani_parameter))
    return model.ActionType(name, tuple(parameters), comment=comment)


JANIModel = t.Union[bytes, str]


JSONObject = t.Mapping[str, t.Any]


def _load_parameter(jani_parameter: JSONObject) -> functions.FunctionParameter:
    _check_fields(jani_parameter, required={"name", "type"})
    return functions.FunctionParameter(
        name=jani_parameter["name"], typ=_type(jani_parameter["type"])
    )


def _load_functions(
    scope: context.Scope, jani_functions: t.Sequence[JSONObject]
) -> None:
    for jani_function in jani_functions:
        _check_fields(jani_function, required={"name", "type", "parameters", "body"})
        scope.define_function(
            name=jani_function["name"],
            parameters=[
                _load_parameter(jani_parameter)
                for jani_parameter in jani_function["parameters"]
            ],
            returns=_type(jani_function["type"]),
            body=_expression(jani_function["body"]),
        )


def load_model(source: JANIModel, *, ignore_properties: bool = False) -> model.Network:
    """
    Constructs a Momba automaton :class:`~momba.model.Network` based on
    the provided JANI model string.

    The flag `ignore_properties` indicates whether the properties of the
    model should be ignored or added to the returned network's modeling
    :class:`~momba.model.Context`.

    Throws a :class:`~momba.jani.JANIError` in case the JANI file cannot be
    loaded.
    """
    if isinstance(source, bytes):
        jani_model = json.loads(source.decode("utf-8"))
    else:
        jani_model = json.loads(source)
    features = {ModelFeature(feature) for feature in jani_model.get("features", [])}
    unsupported_features = features.difference(
        {
            ModelFeature.ARRAYS,
            ModelFeature.DERIVED_OPERATORS,
            ModelFeature.FUNCTIONS,
            ModelFeature.STATE_EXIT_REWARDS,
            ModelFeature.NONDET_SELECTION,
            ModelFeature.TRIGONOMETRIC_FUNCTIONS,
        }
    )
    if unsupported_features:
        raise UnsupportedJANIError(unsupported_features)
    _check_fields(
        jani_model,
        required={"jani-version", "name", "type", "automata", "system"},
        optional={
            "metadata",
            "features",
            "actions",
            "constants",
            "variables",
            "restrict-initial",
            "properties",
            "comment",
            "functions",
        },
        unsupported={"comment"},
    )

    network = model.Network(model.Context(model.ModelType[jani_model["type"].upper()]))
    if "variables" in jani_model:
        for jani_declaration in jani_model["variables"]:
            var_declaration = _variable_declaration(jani_declaration)
            network.ctx.global_scope.add_declaration(var_declaration, validate=False)
    if "constants" in jani_model:
        for jani_declaration in jani_model["constants"]:
            const_declaration = _constant_declaration(jani_declaration)
            network.ctx.global_scope.add_declaration(const_declaration, validate=False)
    for declaration in network.ctx.global_scope.declarations:
        declaration.validate(network.ctx.global_scope)
    if "restrict-initial" in jani_model:
        _check_fields(
            jani_model["restrict-initial"], required={"exp"}, optional={"comment"}
        )
        initial_restriction = _expression(jani_model["restrict-initial"]["exp"])
        network.initial_restriction = initial_restriction
    if "actions" in jani_model:
        for jani_action in jani_model["actions"]:
            network.ctx._add_action_type(_action(jani_action))
    _load_functions(network.ctx.global_scope, jani_model.get("functions", ()))
    for jani_automaton in jani_model["automata"]:
        _check_fields(
            jani_automaton,
            required={"name", "locations", "initial-locations", "edges"},
            optional={
                "variables",
                "functions",
                "restrict-initial",
                "comment",
                "x-momba-anonymous",
            },
        )
        name: t.Optional[str]
        if jani_automaton.get("x-momba-anonymous", False):
            name = None
        else:
            name = jani_automaton["name"]
        automaton = network.ctx.create_automaton(name=name)
        if "variables" in jani_automaton:
            for jani_declaration in jani_automaton["variables"]:
                declaration = _variable_declaration(jani_declaration)
                automaton.scope.add_declaration(declaration, validate=False)
        for declaration in automaton.scope.declarations:
            declaration.validate(automaton.scope)
        locations = {
            jani_location["name"]: _location(jani_location)
            for jani_location in jani_automaton["locations"]
        }
        _load_functions(automaton.scope, jani_automaton.get("functions", ()))
        if "restrict-initial" in jani_automaton:
            _check_fields(
                jani_automaton["restrict-initial"],
                required={"exp"},
                optional={"comment"},
            )
            initial_restriction = _expression(jani_automaton["restrict-initial"]["exp"])
            automaton.initial_restriction = initial_restriction
        for location in locations.values():
            automaton.add_location(location)
        for jani_edge in jani_automaton["edges"]:
            automaton.add_edge(_edge(network.ctx, locations, jani_edge))
        for jani_location in jani_automaton["initial-locations"]:
            automaton.add_location(locations[jani_location], initial=True)
    jani_system = jani_model["system"]
    instances: t.List[model.Instance] = []
    _check_fields(jani_system, required={"elements"}, optional={"syncs", "comment"})
    for element in jani_system["elements"]:
        _check_fields(
            element, required={"automaton"}, optional={"input-enable", "comment"}
        )
        automaton = network.ctx.get_automaton_by_name(element["automaton"])
        input_enable = element.get("input-enable", None)
        instance = automaton.create_instance(
            input_enable=frozenset()
            if input_enable is None
            else frozenset(input_enable)
        )
        network.add_instance(instance)
        instances.append(instance)
    if "syncs" in jani_system:
        for jani_vector in jani_system["syncs"]:
            vector: t.Dict[model.Instance, model.ActionPattern] = {}
            for index, action_name in enumerate(jani_vector["synchronise"]):
                if action_name is None:
                    continue
                instance = instances[index]
                action_type = network.ctx.get_action_type_by_name(action_name)
                vector[instance] = action_type.create_pattern()
            result_pattern = None
            if "result" in jani_vector:
                result_pattern = network.ctx.get_action_type_by_name(
                    jani_vector["result"]
                ).create_pattern()
            network.create_link(vector, result=result_pattern)
    if not ignore_properties:
        for jani_prop in jani_model["properties"]:
            _check_fields(
                jani_prop,
                required={"expression", "name"},
            )
            network.ctx.define_property(
                jani_prop["name"], _property(jani_prop["expression"])
            )
    return network



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

import dataclasses as d
import typing as t

import fractions
import pathlib
import subprocess
import tempfile
import re

from .. import model
from ..analysis import checkers
from ..jani import dump_model

from .errors import ToolError, ToolTimeoutError


Timeout = t.Optional[t.Union[float, int]]
Command = t.Sequence[t.Union[str, pathlib.Path]]


# XXX: is there a better way to do this?
_result_regex = re.compile(
    r"Model checking property \"(?P<prop_name>[^\"]+)\""
    r".*?"
    r"Result \(for initial states\): (?P<prop_value>\d+(\.\d+)?)",
    flags=re.DOTALL,
)


@d.dataclass(frozen=True)
class Output:
    stdout: str
    stderr: str


@d.dataclass(eq=False)
class Toolset:
    """Interface to Storm."""

    executable: t.Union[str, pathlib.Path]
    """Path to the executable."""

    environment: t.Optional[t.Mapping[str, str]] = None
    """Environment variables for execution."""

    def invoke(
        self,
        arguments: t.Iterable[t.Union[str, int, float, pathlib.Path]],
        *,
        timeout: Timeout = None,
    ) -> Output:
        """Runs storm with the provided arguments."""
        command: Command = (
            self.executable,
            *map(str, arguments),
        )
        try:
            process = subprocess.run(
                command,
                env=self.environment,
                timeout=timeout,
                capture_output=True,
            )
        except subprocess.TimeoutExpired as timeout_error:
            raise ToolTimeoutError(
                "timeout expired during invocation of `storm`",
                command=command,
                stdout=timeout_error.stdout,
                stderr=timeout_error.stderr,
            )
        if process.returncode != 0:
            raise ToolError(
                f"`storm` terminated with non-zero returncode {process.returncode}",
                command=command,
                stdout=process.stdout,
                stderr=process.stderr,
                returncode=process.returncode,
            )
        return Output(process.stdout.decode("utf-8"), process.stderr.decode("utf-8"))


@d.dataclass(frozen=True, eq=False)
class StormChecker(checkers.Checker):
    toolset: Toolset

    engine: str = "dd"

    def check(
        self,
        network: model.Network,
        *,
        properties: t.Optional[checkers.Properties] = None,
        property_names: t.Optional[t.Iterable[str]] = None,
    ) -> checkers.Result:
        with tempfile.TemporaryDirectory(prefix="modest") as directory_name:
            input_file = pathlib.Path(directory_name) / "input.jani"
            named_properties: t.Dict[str, model.Expression] = {}
            if properties is None and property_names is None:
                named_properties.update(
                    {
                        definition.name: definition.expression
                        for definition in network.ctx.properties.values()
                    }
                )
            if property_names is not None:
                for name in property_names:
                    named_properties[
                        name
                    ] = network.ctx.get_property_definition_by_name(name).expression
            input_file.write_text(
                dump_model(network, properties=named_properties), encoding="utf-8"
            )
            output = self.toolset.invoke(
                ("--jani", input_file, "--janiproperty", "--engine", self.engine)
            )
        return {
            match.group("prop_name"): fractions.Fraction(match.group("prop_value"))
            for match in _result_regex.finditer(output.stdout)
        }


toolset = Toolset("storm")

checker_sparse = StormChecker(toolset, engine="sparse")
checker_dd = StormChecker(toolset, engine="dd")

checker = checker_sparse


def get_checker(*, accept_license: bool) -> checkers.Checker:
    """Returns an instance of :class:`~momba.analysis.Checker`."""
    try:
        from . import storm_docker

        return storm_docker.checker
    except ImportError:
        return checker



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

"""
Translates a Model into Momba's intermediate representation for state
space exploration.
"""

from __future__ import annotations

import dataclasses as d
import typing as t

import functools
import fractions
import json

from .. import model
from ..model import actions, expressions, operators


if t.TYPE_CHECKING:
    from .explore import Parameters

_JSONObject = t.Dict[str, t.Any]


class CompileError(Exception):
    pass


@functools.singledispatch
def _translate_type(typ: model.Type) -> _JSONObject:
    raise NotImplementedError(f"'_translate_type' not implemented for {type(typ)}")


@_translate_type.register
def _translate_integer_type(typ: model.types.IntegerType) -> _JSONObject:
    return {"type": "INT64"}


@_translate_type.register
def _translate_real_type(typ: model.types.RealType) -> _JSONObject:
    return {"type": "FLOAT64"}


@_translate_type.register
def _translate_bounded_type(typ: model.types.BoundedType) -> _JSONObject:
    return _translate_type(typ.base)


@_translate_type.register
def _translate_bool_type(typ: model.types.BoolType) -> _JSONObject:
    return {"type": "BOOL"}


@_translate_type.register
def _translate_array_type(typ: model.types.ArrayType) -> _JSONObject:
    return {"type": "VECTOR", "element_type": _translate_type(typ.element)}


@d.dataclass(frozen=True)
class VariableDeclaration:
    identifier: str
    typ: model.types.Type
    initial_value: model.Expression
    is_transient: bool


GlobalsTable = t.Mapping[str, VariableDeclaration]
LocalsTable = t.Mapping[model.Instance, t.Mapping[str, VariableDeclaration]]


@d.dataclass
class Declarations:
    globals_table: GlobalsTable = d.field(default_factory=dict)
    locals_table: LocalsTable = d.field(default_factory=dict)

    @property
    def all_declarations(self) -> t.List[VariableDeclaration]:
        all_declarations = list(self.globals_table.values())
        for local_declarations in self.locals_table.values():
            all_declarations.extend(local_declarations.values())
        return all_declarations


@d.dataclass
class _TranslationContext:
    parameters: t.Mapping[str, model.Expression]
    declarations: Declarations
    scope: model.Scope
    instance: t.Optional[model.Instance] = None

    @property
    def instance_table(self) -> t.Optional[t.Mapping[str, VariableDeclaration]]:
        if self.instance is None:
            return None
        return self.declarations.locals_table[self.instance]


def _translate_identifier(identifier: str, ctx: _TranslationContext) -> _JSONObject:
    # TODO: How do we deal with edge and action scopes?
    try:
        declaration = ctx.scope.lookup(identifier)
        if isinstance(declaration, model.ConstantDeclaration):
            try:
                return _translate_expr(ctx.parameters[declaration.identifier], ctx)
            except KeyError:
                pass
            if declaration.value is None:
                raise CompileError(
                    f"No value for constant {declaration.identifier!r} provided."
                )
            return _translate_expr(declaration.value, ctx)
        instance_table = ctx.instance_table
        if instance_table is not None:
            try:
                return {
                    "kind": "NAME",
                    "identifier": instance_table[identifier].identifier,
                }
            except KeyError:
                pass
        return {
            "kind": "NAME",
            "identifier": ctx.declarations.globals_table[identifier].identifier,
        }
    except KeyError or model.errors.UnboundIdentifierError:  # type: ignore
        # We may run in this case due to *anonymous scopes* occuring, for
        # instance, as part of array constructors.
        return {"kind": "NAME", "identifier": identifier}


@functools.singledispatch
def _translate_expr(expr: model.Expression, ctx: _TranslationContext) -> _JSONObject:
    raise NotImplementedError(f"'_translate_expr' not implemented for {type(expr)}")


@_translate_expr.register
def _translate_boolean_constant_expr(
    expr: expressions.BooleanConstant, ctx: _TranslationContext
) -> _JSONObject:
    return {"kind": "CONSTANT", "value": expr.boolean}


@_translate_expr.register
def _translate_integer_constant_expr(
    expr: expressions.IntegerConstant, ctx: _TranslationContext
) -> _JSONObject:
    return {"kind": "CONSTANT", "value": expr.integer}


@_translate_expr.register
def _translate_real_constant_expr(
    expr: expressions.RealConstant, ctx: _TranslationContext
) -> _JSONObject:
    return {"kind": "CONSTANT", "value": expr.as_float}


@_translate_expr.register
def _translate_name_expr(
    expr: expressions.Name, ctx: _TranslationContext
) -> _JSONObject:
    return _translate_identifier(expr.identifier, ctx)


@_translate_expr.register
def _translate_boolean_expr(
    expr: expressions.Boolean, ctx: _TranslationContext
) -> _JSONObject:
    # We flatten trees of boolean expressions. The IR comes with boolean
    # operators taking an arbitrary amount of expressions.
    operator = expr.operator
    operands: t.List[_JSONObject] = []
    stack = [expr.right, expr.left]
    while stack:
        top = stack.pop()
        if isinstance(top, expressions.Boolean) and top.operator == operator:
            stack.append(top.right)
            stack.append(top.left)
        elif isinstance(top, expressions.BooleanConstant):
            if operator is operators.BooleanOperator.AND:
                if not top.boolean:
                    return {"kind": "CONSTANT", "value": False}
            else:
                assert operator is operators.BooleanOperator.OR
                if top.boolean:
                    return {"kind": "CONSTANT", "value": True}
        else:
            operands.append(_translate_expr(top, ctx))
    return {"kind": "BOOLEAN", "operator": operator.name, "operands": operands}


@_translate_expr.register
def _translate_arithmetic_binary_expr(
    expr: expressions.ArithmeticBinary, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "BINARY",
        "operator": expr.operator.name,
        "left": _translate_expr(expr.left, ctx),
        "right": _translate_expr(expr.right, ctx),
    }


_EQUALITY_OPERATOR_MAP = {
    operators.EqualityOperator.EQ: "EQ",
    operators.EqualityOperator.NEQ: "NE",
}


@_translate_expr.register
def _translate_equality_expr(
    expr: expressions.Equality, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "COMPARISON",
        "operator": _EQUALITY_OPERATOR_MAP[expr.operator],
        "left": _translate_expr(expr.left, ctx),
        "right": _translate_expr(expr.right, ctx),
    }


@_translate_expr.register
def _translate_comparison_expr(
    expr: expressions.Comparison, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "COMPARISON",
        "operator": expr.operator.name,
        "left": _translate_expr(expr.left, ctx),
        "right": _translate_expr(expr.right, ctx),
    }


@_translate_expr.register
def _translate_conditional_expr(
    expr: expressions.Conditional, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "CONDITIONAL",
        "condition": _translate_expr(expr.condition, ctx),
        "consequence": _translate_expr(expr.consequence, ctx),
        "alternative": _translate_expr(expr.alternative, ctx),
    }


@_translate_expr.register
def _translate_arithmetic_unary_expr(
    expr: expressions.ArithmeticUnary, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "UNARY",
        "operator": expr.operator.name,
        "operand": _translate_expr(expr.operand, ctx),
    }


@_translate_expr.register
def _translate_not_expr(expr: expressions.Not, ctx: _TranslationContext) -> _JSONObject:
    return {
        "kind": "UNARY",
        "operator": "NOT",
        "operand": _translate_expr(expr.operand, ctx),
    }


@_translate_expr.register
def _translate_array_access_expr(
    expr: expressions.ArrayAccess, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "INDEX",
        "vector": _translate_expr(expr.array, ctx),
        "index": _translate_expr(expr.index, ctx),
    }


@_translate_expr.register
def _translate_array_value_expr(
    expr: expressions.ArrayValue, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "VECTOR",
        "elements": [_translate_expr(element, ctx) for element in expr.elements],
    }


@_translate_expr.register
def _translate_array_constructor_expr(
    expr: expressions.ArrayConstructor, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "COMPREHENSION",
        "variable": expr.variable,
        "length": _translate_expr(expr.length, ctx),
        "element": _translate_expr(expr.expression, ctx),
    }


@_translate_expr.register
def _translate_trigonometric_expr(
    expr: expressions.Trigonometric, ctx: _TranslationContext
) -> _JSONObject:
    return {
        "kind": "TRIGONOMETRIC",
        "function": expr.operator.name,
        "operand": _translate_expr(expr.operand, ctx),
    }


def _compute_instance_names(network: model.Network) -> t.Mapping[model.Instance, str]:
    instance_names: t.Dict[model.Instance, str] = {}
    for number, instance in enumerate(network.instances):
        instance_names[instance] = f"{number}_{instance.automaton.name}"
    return instance_names


def _extract_initial_value(declaration: model.VariableDeclaration) -> model.Expression:
    if declaration.initial_value is None:
        raise CompileError(
            f"No initial value for variable {declaration.identifier!r} provided."
        )
    return declaration.initial_value


def _compute_declarations_with_prefix(
    prefix: str, declarations: t.Iterable[model.VariableDeclaration]
) -> t.Mapping[str, VariableDeclaration]:
    return {
        declaration.identifier: VariableDeclaration(
            f"{prefix}{declaration.identifier}",
            declaration.typ,
            _extract_initial_value(declaration),
            declaration.is_transient or False,
        )
        for declaration in declarations
    }


def _compute_declarations(
    network: model.Network, instance_names: t.Mapping[model.Instance, str]
) -> Declarations:
    return Declarations(
        _compute_declarations_with_prefix(
            "global_", network.ctx.global_scope.variable_declarations
        ),
        {
            instance: _compute_declarations_with_prefix(
                f"local_{instance_name}_",
                instance.automaton.scope.variable_declarations,
            )
            for instance, instance_name in instance_names.items()
        },
    )


def _insert_variable_declaration(
    global_variables: _JSONObject, declaration: VariableDeclaration
) -> None:
    identifier = declaration.identifier
    typ = _translate_type(declaration.typ)
    global_variables[identifier] = typ


def _compute_global_variables(declarations: Declarations) -> _JSONObject:
    global_variables: _JSONObject = {}
    for variable_declaration in declarations.globals_table.values():
        if variable_declaration.is_transient:
            continue
        elif variable_declaration.typ == model.types.CLOCK:
            continue
        _insert_variable_declaration(global_variables, variable_declaration)
    for instance_declarations in declarations.locals_table.values():
        for variable_declaration in instance_declarations.values():
            if variable_declaration.is_transient:
                continue
            elif variable_declaration.typ == model.types.CLOCK:
                continue
            _insert_variable_declaration(global_variables, variable_declaration)
    return global_variables


def _compute_action_types(network: model.Network) -> _JSONObject:
    return {
        action_type.label: [
            _translate_type(parameter.typ) for parameter in action_type.parameters
        ]
        for action_type in network.ctx.action_types.values()
    }


def _translate_pattern_argument(
    argument: model.ActionArgument, ctx: _TranslationContext
) -> _JSONObject:
    if isinstance(argument, model.ReadArgument):
        return {"direction": "READ", "identifier": argument.identifier}
    else:
        assert isinstance(
            argument, model.WriteArgument
        ), "guard arguments are not supported"
        return {
            "direction": "WRITE",
            "value": _translate_expr(argument.expression, ctx),
        }


def _translate_action_pattern(
    action_pattern: t.Optional[model.ActionPattern],
    ctx: _TranslationContext,
) -> _JSONObject:
    if action_pattern is None:
        return {"kind": "SILENT"}
    else:
        return {
            "kind": "LABELED",
            "label": action_pattern.action_type.label,
            "arguments": [
                _translate_pattern_argument(argument, ctx)
                for argument in action_pattern.arguments
            ],
        }


def _contains_clock_identifier(
    expression: expressions.Expression, scope: model.Scope
) -> bool:
    for used_name in expression.used_names:
        if scope.lookup(used_name.identifier).typ == model.types.CLOCK:
            return True
    return False


@d.dataclass(frozen=True)
class ExtractedConstraints:
    conjuncts: t.List[expressions.Expression]
    constraints: t.List[_JSONObject]


def _extract_constraints(
    expr: model.Expression,
    ctx: _TranslationContext,
) -> ExtractedConstraints:
    constraints: t.List[_JSONObject] = []
    conjuncts: t.List[expressions.Expression] = []
    pending: t.List[expressions.Expression] = [expr]
    while pending:
        head = pending.pop()
        if isinstance(head, expressions.Boolean):
            if head.operator is operators.BooleanOperator.AND:
                pending.append(head.left)
                pending.append(head.right)
            else:
                conjuncts.append(head)
        elif isinstance(head, expressions.Comparison):
            if _contains_clock_identifier(head.left, ctx.scope):
                difference = head.left
                operator = head.operator
                bound_expression = head.right
            elif _contains_clock_identifier(head.right, ctx.scope):
                difference = head.right
                operator = head.operator.swap()
                bound_expression = head.left
            else:
                conjuncts.append(head)
                continue
            assert not _contains_clock_identifier(bound_expression, ctx.scope)
            left: _JSONObject
            right: _JSONObject
            if isinstance(difference, expressions.Name):
                left = {
                    "kind": "VARIABLE",
                    "identifier": _translate_identifier(difference.identifier, ctx)[
                        "identifier"
                    ],
                }
                right = {"kind": "ZERO"}
            else:
                assert (
                    isinstance(difference, expressions.ArithmeticBinary)
                    and difference.operator is operators.ArithmeticBinaryOperator.SUB
                    and isinstance(difference.left, expressions.Name)
                    and isinstance(difference.right, expressions.Name)
                )
                left = {
                    "kind": "VARIABLE",
                    "identifier": _translate_identifier(
                        difference.left.identifier, ctx
                    )["identifier"],
                }
                right = {
                    "kind": "VARIABLE",
                    "identifier": _translate_identifier(
                        difference.right.identifier, ctx
                    )["identifier"],
                }
            bound = _translate_expr(bound_expression, ctx)
            if operator.is_less:
                constraints.append(
                    {
                        "left": left,
                        "right": right,
                        "is_strict": operator.is_strict,
                        "bound": bound,
                    }
                )
            else:
                assert operator.is_greater
                constraints.append(
                    {
                        "left": right,
                        "right": left,
                        "is_strict": operator.is_strict,
                        "bound": {
                            "kind": "BINARY",
                            "operator": "SUB",
                            "left": {"kind": "CONSTANT", "value": 0},
                            "right": bound,
                        },
                    }
                )
        else:
            conjuncts.append(head)
    return ExtractedConstraints(conjuncts, constraints)


def _compute_location_names(instance: model.Instance) -> t.Mapping[model.Location, str]:
    return {
        location: f"{location_index}_{location.name or ''}"
        for location_index, location in enumerate(instance.automaton.locations)
    }


def _extract_invariant(
    location: model.Location, ctx: _TranslationContext
) -> t.List[_JSONObject]:
    if location.progress_invariant is None:
        return []
    extracted_constraints = _extract_constraints(location.progress_invariant, ctx)
    assert (
        not extracted_constraints.conjuncts
    ), "invariant must be a conjunction of clock constraints"
    return extracted_constraints.constraints


def _translate_instance(
    instance: model.Instance,
    declarations: Declarations,
    location_names: t.Mapping[model.Location, str],
    parameters: t.Mapping[str, model.Expression],
) -> _JSONObject:
    parameters = dict(parameters)

    for parameter_name, parameter_value in zip(
        instance.automaton.parameters, instance.arguments
    ):
        parameters[parameter_name] = parameter_value

    ctx = _TranslationContext(
        parameters, declarations, instance.automaton.scope, instance
    )
    locations: t.Mapping[str, _JSONObject] = {
        location_name: {
            "invariant": _extract_invariant(location, ctx),
            "edges": [],
        }
        for location, location_name in location_names.items()
    }

    for number, edge in enumerate(instance.automaton.edges):
        outgoing = locations[location_names[edge.location]]["edges"]
        action = _translate_action_pattern(edge.action_pattern, ctx)

        guard: _JSONObject
        if edge.guard is None:
            guard = {
                "boolean_condition": {"kind": "CONSTANT", "value": True},
                "clock_constraints": [],
            }
        else:
            extracted_constraints = _extract_constraints(edge.guard, ctx)
            guard = {
                "boolean_condition": _translate_expr(
                    expressions.logic_all(*extracted_constraints.conjuncts), ctx
                ),
                "clock_constraints": extracted_constraints.constraints,
            }

        destinations: t.List[_JSONObject] = []

        edge_ctx = _TranslationContext(
            parameters,
            declarations,
            edge.create_edge_scope(instance.automaton.scope),
            instance,
        )

        for destination in edge.destinations:
            probability: _JSONObject
            if destination.probability is None:
                probability = {"kind": "CONSTANT", "value": 1.0}
            else:
                probability = _translate_expr(destination.probability, edge_ctx)

            assignments: t.List[_JSONObject] = []

            reset_clocks: t.List[_JSONObject] = []

            for assignment in destination.assignments:
                assert isinstance(assignment.target, expressions.Name)
                declaration = edge_ctx.scope.lookup(assignment.target.identifier)
                if declaration.typ == model.types.CLOCK:
                    assert assignment.index == 0
                    assert isinstance(assignment.value, expressions.IntegerConstant)
                    reset_clocks.append(
                        {
                            "kind": "VARIABLE",
                            "identifier": _translate_identifier(
                                assignment.target.identifier, ctx
                            )["identifier"],
                        }
                    )
                else:
                    assignments.append(
                        {
                            "target": _translate_identifier(
                                assignment.target.identifier, edge_ctx
                            ),
                            "value": _translate_expr(assignment.value, edge_ctx),
                            "index": assignment.index,
                        }
                    )

            destinations.append(
                {
                    "location": location_names[destination.location],
                    "probability": probability,
                    "assignments": assignments,
                    "reset": reset_clocks,
                }
            )

        outgoing.append(
            {
                "number": number,
                "pattern": action,
                "guard": guard,
                "destinations": destinations,
                "observations": [
                    {
                        "label": observation.action_type.label,
                        "arguments": [
                            _translate_expr(argument, edge_ctx)
                            for argument in observation.arguments
                        ],
                        "probability": _translate_expr(
                            observation.probability
                            or expressions.RealConstant(fractions.Fraction(1.0)),
                            edge_ctx,
                        ),
                    }
                    for observation in edge.observation
                ],
            }
        )

    return {"locations": locations}


def _translate_link(
    link: model.Link, instance_names: t.Mapping[model.Instance, str]
) -> _JSONObject:
    slots: t.List[str] = []

    def _translate_arguments(
        arguments: t.Iterable[model.ActionArgument],
    ) -> t.List[str]:
        slot_vector: t.List[str] = []
        for argument in arguments:
            assert isinstance(argument, actions.GuardArgument)
            if argument.identifier not in slots:
                slots.append(argument.identifier)
            slot_vector.append(argument.identifier)
        return slot_vector

    assert link.condition is None
    result: _JSONObject
    if link.result is None:
        result = {"kind": "SILENT"}
    else:
        result = {
            "kind": "LABELED",
            "action_type": link.result.action_type.label,
            "arguments": _translate_arguments(link.result.arguments),
        }
    vector: _JSONObject = {}
    for instance, pattern in link.vector.items():
        vector[instance_names[instance]] = {
            "action_type": pattern.action_type.label,
            "arguments": _translate_arguments(pattern.arguments),
        }

    return {"slots": slots, "vector": vector, "result": result}


def _extract_constant_value(
    expr: expressions.Expression,
    parameters: t.Mapping[str, model.Expression],
) -> t.Any:
    if isinstance(expr, expressions.IntegerConstant):
        return expr.integer
    elif isinstance(expr, expressions.BooleanConstant):
        return expr.boolean
    elif isinstance(expr, expressions.RealConstant):
        return expr.as_float
    elif isinstance(expr, expressions.ArrayValue):
        return [
            _extract_constant_value(element, parameters) for element in expr.elements
        ]
    elif isinstance(expr, expressions.Name):
        return _extract_constant_value(parameters[expr.identifier], parameters)
    elif isinstance(expr, expressions.BinaryExpression) and hasattr(
        expr.operator, "native_function"
    ):
        left = _extract_constant_value(expr.left, parameters)
        right = _extract_constant_value(expr.right, parameters)
        return expr.operator.native_function(left, right)
    else:
        raise CompileError(f"Unable to extract constant value from {expr}.")


def _update_initial_values(
    initial_values: _JSONObject,
    clock_constraints: t.List[_JSONObject],
    declarations: t.Iterable[VariableDeclaration],
    parameters: t.Mapping[str, model.Expression],
) -> None:
    for declaration in declarations:
        initial_value = declaration.initial_value
        assert initial_value is not None
        value = _extract_constant_value(initial_value, parameters)
        if declaration.is_transient:
            continue
        elif declaration.typ == model.types.CLOCK:
            assert isinstance(value, (int, float))
            clock_constraints.append(
                {
                    "left": {"kind": "VARIABLE", "identifier": declaration.identifier},
                    "right": {"kind": "ZERO"},
                    "is_strict": False,
                    "bound": {"kind": "CONSTANT", "value": value},
                }
            )
            clock_constraints.append(
                {
                    "left": {"kind": "ZERO"},
                    "right": {"kind": "VARIABLE", "identifier": declaration.identifier},
                    "is_strict": False,
                    "bound": {"kind": "CONSTANT", "value": -value},
                }
            )
        else:
            initial_values[declaration.identifier] = value


def _translate_initial_states(
    network: model.Network,
    instance_names: t.Mapping[model.Instance, str],
    instance_to_location_names: t.Mapping[
        model.Instance, t.Mapping[model.Location, str]
    ],
    declarations: Declarations,
    parameters: t.Mapping[str, model.Expression],
) -> t.List[_JSONObject]:
    initial_locations: t.Dict[str, str] = {}
    for instance, instance_name in instance_names.items():
        # TODO: support more than one initial location
        (initial_location,) = instance.automaton.initial_locations
        location_name = instance_to_location_names[instance][initial_location]
        initial_locations[instance_name] = location_name
    initial_values: _JSONObject = {}
    clock_constraints: t.List[_JSONObject] = []
    _update_initial_values(
        initial_values,
        clock_constraints,
        declarations.globals_table.values(),
        parameters,
    )
    for variable_declarations in declarations.locals_table.values():
        _update_initial_values(
            initial_values,
            clock_constraints,
            variable_declarations.values(),
            parameters,
        )
    # for declaration in declarations.all_declarations:
    #     if declaration.typ != model.types.CLOCK:
    #         continue
    #     for other_declaration in declarations.all_declarations:
    #         if (
    #             declaration == other_declaration
    #             or other_declaration.typ != model.types.CLOCK
    #         ):
    #             continue
    #         clock_constraints.append(
    #             {
    #                 "left": {"kind": "VARIABLE", "identifier": declaration.identifier},
    #                 "right": {"kind": "VARIABLE", "identifier": other_declaration.identifier},
    #                 "is_strict": False,
    #                 "bound": {"kind": "CONSTANT", "value": },
    #             }
    #         )
    return [
        {
            "locations": initial_locations,
            "values": initial_values,
            "zone": clock_constraints,
        }
    ]


def _translate_network(
    network: model.Network,
    instance_names: t.Mapping[model.Instance, str],
    instance_to_location_names: t.Mapping[
        model.Instance, t.Mapping[model.Location, str]
    ],
    declarations: Declarations,
    parameters: t.Mapping[str, model.Expression],
    global_clock: bool,
) -> str:
    ctx = _TranslationContext(parameters, declarations, network.ctx.global_scope, None)
    return json.dumps(
        {
            "declarations": {
                "global_variables": _compute_global_variables(declarations),
                "transient_variables": {
                    declaration.identifier: _translate_expr(
                        declaration.initial_value, ctx
                    )
                    for declaration in declarations.all_declarations
                    if declaration.is_transient
                },
                "clock_variables": [
                    declaration.identifier
                    for declaration in declarations.all_declarations
                    if declaration.typ == model.types.CLOCK
                ]
                + (["global_clock"] if global_clock else []),
                "action_labels": _compute_action_types(network),
            },
            "automata": {
                instance_name: _translate_instance(
                    instance,
                    declarations,
                    instance_to_location_names[instance],
                    parameters,
                )
                for instance, instance_name in instance_names.items()
            },
            "links": [_translate_link(link, instance_names) for link in network.links],
            "initial_states": _translate_initial_states(
                network,
                instance_names,
                instance_to_location_names,
                declarations,
                parameters,
            ),
        },
        indent=2,
    )


@d.dataclass(frozen=True)
class Translation:
    network: model.Network
    parameters: t.Mapping[str, model.Expression]
    json_network: str
    instance_names: t.Mapping[model.Instance, str]
    declarations: Declarations
    instance_to_location_names: t.Mapping[
        model.Instance, t.Mapping[model.Location, str]
    ]
    instance_vector: t.Tuple[model.Instance, ...]

    @functools.cached_property
    def instance_name_to_instance(self) -> t.Mapping[str, model.Instance]:
        return {name: instance for instance, name in self.instance_names.items()}

    @functools.cached_property
    def reversed_instance_to_location_names(
        self,
    ) -> t.Mapping[model.Instance, t.Mapping[str, model.Location]]:
        return {
            instance: {name: location for location, name in mapping.items()}
            for instance, mapping in self.instance_to_location_names.items()
        }

    def translate_global_expression(self, expr: model.Expression) -> str:
        ctx = _TranslationContext(
            self.parameters,
            self.declarations,
            self.network.ctx.global_scope,
            None,
        )
        return json.dumps(_translate_expr(expr, ctx))


def translate_network(
    network: model.Network, *, parameters: Parameters = None, global_clock: bool = False
) -> Translation:
    instance_names = _compute_instance_names(network)
    declarations = _compute_declarations(network, instance_names)
    instance_to_location_names = {
        instance: _compute_location_names(instance) for instance in network.instances
    }
    ctx_parameters = {
        name: expressions.ensure_expr(expr) for name, expr in (parameters or {}).items()
    }
    return Translation(
        network,
        ctx_parameters,
        _translate_network(
            network,
            instance_names,
            instance_to_location_names,
            declarations,
            ctx_parameters,
            global_clock,
        ),
        instance_names,
        declarations,
        instance_to_location_names,
        # HACK: this relies on dictionaries being ordered on both the Python and Rust side
        instance_vector=tuple(instance_names.keys()),
    )



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import collections

from mxu.maps import FrozenMap

from . import actions, errors, expressions, observations, types

if t.TYPE_CHECKING:
    from . import context


Annotation = t.Mapping[str, t.Union[int, str, float]]


@d.dataclass(frozen=True, eq=False)
class Instance:
    """
    Represents an automaton instance.

    Attributes
    ----------
    automaton:
        The instaniated automaton.
    arguments:
        The arguments passed to the parameters of the automaton.
    input_enabled:
        The set of action types for which the instance is input enabled.
    comment:
        An optional comment describing the instance.
    """

    automaton: Automaton
    arguments: t.Tuple[expressions.Expression, ...] = ()
    input_enable: t.FrozenSet[actions.ActionType] = frozenset()
    comment: t.Optional[str] = None

    def __post_init__(self) -> None:
        if len(self.arguments) != len(self.automaton.parameters):
            raise errors.ModelingError("invalid number of arguments for automaton")


ProgressInvariant = t.Optional[expressions.Expression]
TransientValues = t.AbstractSet["Assignment"]


@d.dataclass(frozen=True, eq=False)
class Location:
    """
    Represents a location of an automaton.

    Attributes:
        name:
            The optional name of the location.
        progress_invariant:
            An optional expression `progress_invariant` specifies the invariant
            subject to which the time may progress in the location. As per the
            JANI specification, the invariant is only allowed for models using
            real-valued clocks (see :class:`~momba.model.ModelType`).
        transient_values:
            A set of assignments for transient variables. These assignments
            define the values of the transient variables when in the location.
    """

    name: t.Optional[str] = None

    progress_invariant: ProgressInvariant = None
    transient_values: TransientValues = frozenset()

    def validate(self, automaton: Automaton) -> None:
        """
        Validates the location for the given automaton.

        Raises :class:`~errors.ModelingError` if the location is invalid
        to add to the automaton according to the JANI specification. For
        instance, if the location has a `progress_invariant` but the model
        type does not support clocks.
        """
        scope = automaton.scope
        if self.progress_invariant is not None:
            if not scope.ctx.model_type.uses_clocks:
                raise errors.ModelingError(
                    f"location invariant is not allowed for model type {scope.ctx.model_type}"
                )
            if scope.get_type(self.progress_invariant) != types.BOOL:
                raise errors.InvalidTypeError(
                    f"type of invariant in location {self} is not `types.BOOL`"
                )
        if self.transient_values:
            if not are_compatible(self.transient_values):
                raise errors.IncompatibleAssignmentsError(
                    f"incompatible assignments for transient values in location {self}"
                )
            for assignment in self.transient_values:
                if assignment.index != 0:
                    raise errors.ModelingError(
                        f"index of assignment {assignment} to transient variable is non-zero"
                    )
                assignment.validate(scope)


@d.dataclass(frozen=True)
class Assignment:
    """
    Represents an assignment.

    Attributes
    ----------
    target:
        The target of the assignment.
    value:
        The value to assign.
    index:
        The index of the assignment.
    """

    target: expressions.Expression
    value: expressions.Expression
    index: int = 0

    def validate(self, scope: context.Scope) -> None:
        """
        Validates the assignment in the given scope.

        Raises :class:`~errors.ModelingError` if the target is not
        assignable from the value type.
        """
        target_type = self.target.infer_target_type(scope)
        value_type = scope.get_type(self.value)
        if not target_type.is_assignable_from(value_type):
            raise errors.InvalidTypeError(
                f"cannot assign {value_type} to {target_type}"
            )


def are_compatible(assignments: t.Iterable[Assignment]) -> bool:
    """
    Checks whether the given assignments are compatible according
    to the JANI specification.
    """
    groups: t.DefaultDict[int, t.Set[expressions.Expression]] = collections.defaultdict(
        set
    )
    for assignment in assignments:
        target = assignment.target
        if target in groups[assignment.index]:
            return False
        groups[assignment.index].add(target)
    return True


@d.dataclass(frozen=True)
class Destination:
    """
    Represents a destination of an edge.

    Attributes
    ----------
    location:
        The target location of the destination.
    probability:
        An optional expression for the probability of the destination.
    assignments:
        A set of assignments to be executed when going to the destination.
    """

    location: Location
    probability: t.Optional[expressions.Expression] = None
    assignments: t.Tuple[Assignment, ...] = ()

    def _validate(self, automaton: Automaton, scope: context.Scope) -> None:
        self.location.validate(automaton)
        if self.location not in automaton.locations:
            raise errors.ModelingError(
                f"source location of edge {self} is not a location of the automaton {automaton}"
            )
        if not are_compatible(self.assignments):
            raise errors.IncompatibleAssignmentsError(
                f"assignments on edge {self} are not compatible"
            )
        if self.probability is not None:
            if not scope.get_type(self.probability).is_numeric:
                raise errors.InvalidTypeError("probability value must be numeric")
        for assignment in self.assignments:
            assignment.validate(scope)


@d.dataclass(frozen=True)
class Edge:
    """
    Represents an edge of an automaton.

    Attributes
    ----------
    location:
        The source location of the edge.
    destinations:
        The destinations of the edge.
    action_pattern:
        The optional action pattern of the edge.
    guard:
        The optional guard of the edge.
    rate:
        The optional rate of the edge.
    annotation:
        An optional annotation of the edge.
    """

    location: Location
    destinations: t.Tuple[Destination, ...]
    action_pattern: t.Optional[actions.ActionPattern] = None
    guard: t.Optional[expressions.Expression] = None
    rate: t.Optional[expressions.Expression] = None
    annotation: t.Optional[Annotation] = None
    observation: t.FrozenSet[observations.Observation] = frozenset()

    def create_edge_scope(self, parent: context.Scope) -> context.Scope:
        """
        Creates an *edge scope* with the given parent scope.

        .. warning::
            Used for *value passing* an experimental Momba feature. Value
            passing is not part of the official JANI specification.
        """
        scope = parent.create_child_scope()
        if self.action_pattern is not None:
            self.action_pattern.declare_in(scope)
        return scope

    def validate(self, automaton: Automaton) -> None:
        """
        Validates the edge for the given automaton.

        Raises :class:`~errors.ModelingError` if the edge is invalid to add
        to the automaton according to the JANI specification. For instance, if
        source location is not a location of the automaton.
        """
        scope = automaton.scope
        self.location.validate(automaton)
        if self.location not in automaton.locations:
            raise errors.ModelingError(
                f"source location of edge {self} is not a location of the automaton {automaton}"
            )
        if self.rate is not None and not scope.get_type(self.rate).is_numeric:
            raise errors.InvalidTypeError(f"type of rate on edge {self} is not numeric")
        edge_scope = self.create_edge_scope(scope)
        if self.guard is not None and edge_scope.get_type(self.guard) != types.BOOL:
            raise errors.InvalidTypeError(
                f"type of guard on edge {self} is not `types.BOOL`"
            )
        for destination in self.destinations:
            destination._validate(automaton, edge_scope)


class Automaton:
    """
    Represents an automaton.

    Attributes
    ----------
    ctx:
        The :class:`Context` associated with the automaton.
    scope:
        The local :class:`Scope` of the automaton.
    name:
        The optional name of the automaton.
    """

    ctx: context.Context
    scope: context.Scope

    name: t.Optional[str]

    _parameters: t.List[str]

    _locations: t.Set[Location]
    _initial_locations: t.Set[Location]
    _initial_restriction: t.Optional[expressions.Expression]
    _edges: t.List[Edge]
    _incoming_edges: t.DefaultDict[Location, t.Set[Edge]]
    _outgoing_edges: t.DefaultDict[Location, t.Set[Edge]]

    def __init__(self, ctx: context.Context, *, name: t.Optional[str] = None) -> None:
        self.ctx = ctx
        self.scope = self.ctx.global_scope.create_child_scope()
        self.name = name
        self._parameters = []
        self._locations = set()
        self._initial_locations = set()
        self._initial_restriction = None
        self._edges = []
        self._incoming_edges = collections.defaultdict(set)
        self._outgoing_edges = collections.defaultdict(set)

    def __repr__(self) -> str:
        return f"<Automaton name={self.name!r} at 0x{id(self):x}>"

    @property
    def locations(self) -> t.AbstractSet[Location]:
        """
        The set of locations of the automaton.
        """
        return self._locations

    @property
    def initial_locations(self) -> t.AbstractSet[Location]:
        """
        The set of initial locations of the automaton.
        """
        return self._initial_locations

    @property
    def edges(self) -> t.Sequence[Edge]:
        """
        The set of edges of the automaton.
        """
        return self._edges

    @property
    def initial_restriction(self) -> t.Optional[expressions.Expression]:
        """
        An optional restriction to be satisfied by initial states.

        This property can be set *only once* per automaton. The expression
        has to be boolean.

        Raises :class:`~errors.ModelingError` when set twice or to a
        non-boolean expression.
        """
        return self._initial_restriction

    @initial_restriction.setter
    def initial_restriction(self, initial_restriction: expressions.Expression) -> None:
        if self._initial_restriction is not None:
            raise errors.ModelingError(
                "restriction of initial environment has already been set"
            )
        if self.scope.get_type(initial_restriction) != types.BOOL:
            raise errors.ModelingError(
                "restriction of initial valuations must have type `types.BOOL`"
            )
        self._initial_restriction = initial_restriction

    @property
    def parameters(self) -> t.Sequence[str]:
        """
        The sequence of parameters of the automaton.
        """
        return self._parameters

    def add_location(self, location: Location, *, initial: bool = False) -> None:
        """
        Adds a location to the automaton.

        The flag `initial` specifies whether the location
        is an initial location.

        Raises :class:`~errors.ModelingError` when the location cannot be added
        to the automaton. See the method :meth:`Location.validate` for details.
        """
        location.validate(self)
        self._locations.add(location)
        if initial:
            self._initial_locations.add(location)

    def create_location(
        self,
        name: t.Optional[str] = None,
        *,
        progress_invariant: t.Optional[expressions.Expression] = None,
        transient_values: t.AbstractSet[Assignment] = frozenset(),
        initial: bool = False,
    ) -> Location:
        """
        Creates a location with the given name and adds it to the automaton.

        The optional expression `progress_invariant` specifies the invariant
        subject to which the time may progress in the location. As per the
        JANI specification, the invariant is only allowed for models using
        real-valued clocks (see :class:`~momba.model.ModelType`).

        The parameter `transient_values` should be a set of assignments for
        transient variables. These assignments define the values of the
        transient variables when in the location.

        The flag `initial` specifies whether the location
        is an initial location.

        Raises the same exceptions as :meth:`add_location`.
        """
        location = Location(name, progress_invariant, transient_values)
        self.add_location(location, initial=initial)
        return location

    def add_edge(self, edge: Edge) -> None:
        """
        Adds an edge to the automaton.

        Raises :class:`~errors.ModelingError` when the edge cannot be added
        to the automaton. See the method :meth:`Edge.validate` for details.
        """
        edge.validate(self)
        self._edges.append(edge)
        self._locations.add(edge.location)
        self._outgoing_edges[edge.location].add(edge)
        for destination in edge.destinations:
            self._locations.add(destination.location)
            self._incoming_edges[destination.location].add(edge)

    def create_edge(
        self,
        source: Location,
        destinations: t.Iterable[Destination],
        *,
        action_pattern: t.Optional[actions.ActionPattern] = None,
        guard: t.Optional[expressions.Expression] = None,
        rate: t.Optional[expressions.Expression] = None,
        annotation: t.Optional[Annotation] = None,
        observations: t.AbstractSet[observations.Observation] = frozenset(),
    ) -> None:
        """
        Creates an edge and adds it to the automaton.

        The parameters are passed to :class:`Edge`.

        Raises the same exceptions as :meth:`add_edge`.
        """
        edge = Edge(
            source,
            tuple(destinations),
            action_pattern,
            guard,
            rate,
            FrozenMap(annotation),
            frozenset(observations),
        )
        self.add_edge(edge)

    def get_incoming_edges(self, location: Location) -> t.AbstractSet[Edge]:
        """
        Returns the set of outgoing edges from the given location.
        """
        return self._incoming_edges[location]

    def get_outgoing_edges(self, location: Location) -> t.AbstractSet[Edge]:
        """
        Returns the set of incoming edges to the given location.
        """
        return self._outgoing_edges[location]

    def declare_variable(
        self,
        name: str,
        typ: types.Type,
        *,
        is_transient: t.Optional[bool] = None,
        initial_value: t.Optional[expressions.Expression] = None,
    ) -> None:
        """
        Declares a variable in the local scope of the automaton.

        Passes the parameters to :meth:`Scope.declare_variable`.
        """
        self.scope.declare_variable(
            name, typ, is_transient=is_transient, initial_value=initial_value
        )

    def declare_parameter(
        self,
        name: str,
        typ: types.Type,
        *,
        default_value: t.Optional[expressions.Expression] = None,
    ) -> None:
        """
        Declarse a parameter of the automaton.

        Passes the parameters to :meth:`Scope.declare_constant` where
        `value` is `default_value`.
        """
        self.scope.declare_constant(name, typ, value=default_value)
        self._parameters.append(name)

    def create_instance(
        self,
        *,
        arguments: t.Sequence[expressions.ValueOrExpression] = (),
        input_enable: t.AbstractSet[actions.ActionType] = frozenset(),
        comment: t.Optional[str] = None,
    ) -> Instance:
        """
        Creates an instance of the automaton for composition.

        Passes the parameters to :class:`Instance`.
        """
        return Instance(
            self,
            arguments=tuple(map(expressions.ensure_expr, arguments)),
            input_enable=frozenset(input_enable),
        )


Assignments = t.Union[
    t.Iterable["Assignment"], t.Mapping[str, "expressions.Expression"]
]


def create_destination(
    location: Location,
    *,
    probability: t.Optional[expressions.Expression] = None,
    assignments: Assignments = (),
) -> Destination:
    """
    Creates a destination with the given target location.

    This is a convenience function for creating destinations where
    assignments can be provied as a mapping. We recommend using it
    instead of creating :class:`Destination` objects directly.
    """
    if isinstance(assignments, t.Mapping):
        return Destination(
            location,
            probability,
            assignments=tuple(
                Assignment(expressions.Name(name), value)
                for name, value in assignments.items()
            ),
        )
    else:
        return Destination(location, probability, tuple(assignments))



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import abc


StateVector = t.Sequence[float]
AvailableVector = t.Sequence[bool]


@d.dataclass(frozen=True)
class Destination:
    state: StateVector
    reward: float
    probability: float


@d.dataclass(frozen=True)
class Transition:
    action: int
    destinations: t.Sequence[Destination]


class Explorer(abc.ABC):
    """State space explorer for training decision agents."""

    @property
    @abc.abstractmethod
    def num_actions(self) -> int:
        """The number of *actions*."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def num_features(self) -> int:
        """The number of features of the state vector."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def has_terminated(self) -> bool:
        """Indicates whether the explorer is in a terminal state."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def state_vector(self) -> StateVector:
        """The state vector of the current explorer state."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def available_actions(self) -> AvailableVector:
        """A boolean vector indicating which actions are available."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def available_transitions(self) -> t.Sequence[Transition]:
        """A sequence of available transitions."""
        raise NotImplementedError()

    @abc.abstractmethod
    def step(self, action: int) -> float:
        """
        Takes a step with the given action and returns the reward.

        Precondition: The explorer must not be in a terminal state.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def reset(self) -> None:
        """Resets the explorer to the initial state."""
        raise NotImplementedError()

    @abc.abstractmethod
    def fork(self) -> Explorer:
        """Forks the explorer with the current state."""
        raise NotImplementedError()


class Oracle(t.Protocol):
    """An *oracle* selects an action based on the state and the available actions."""

    def __call__(self, state: StateVector, available: AvailableVector) -> int:
        pass



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

from .dump_model import dump_model, ModelFeature
from .load_model import load_model, InvalidJANIError, UnsupportedJANIError, JANIError


__all__ = [
    "dump_model",
    "load_model",
    "InvalidJANIError",
    "UnsupportedJANIError",
    "JANIError",
    "ModelFeature",
]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

from . import errors, expressions, operators, observations, properties, types

from .actions import (
    ActionType,
    ActionPattern,
    ActionParameter,
    ActionArgument,
    ReadArgument,
    WriteArgument,
    GuardArgument,
)

from .automata import (
    Assignment,
    Instance,
    Automaton,
    Location,
    Destination,
    Edge,
    create_destination,
)

from .context import (
    ModelType,
    IdentifierDeclaration,
    VariableDeclaration,
    ConstantDeclaration,
    PropertyDefinition,
    Scope,
    Context,
)

from .distributions import DistributionType

from .expressions import Expression, ensure_expr

from .functions import FunctionDefinition

from .networks import Link, Network

from .types import Type


__all__ = [
    "errors",
    "expressions",
    "operators",
    "observations",
    "properties",
    "types",
    # from `.actions`
    "ActionType",
    "ActionPattern",
    "ActionParameter",
    "ActionArgument",
    "ReadArgument",
    "WriteArgument",
    "GuardArgument",
    # from `.automata`
    "Assignment",
    "Instance",
    "Automaton",
    "Location",
    "Destination",
    "Edge",
    "create_destination",
    # from `.context`
    "ModelType",
    "IdentifierDeclaration",
    "VariableDeclaration",
    "ConstantDeclaration",
    "PropertyDefinition",
    "Scope",
    "Context",
    # from `.distributions`
    "DistributionType",
    # from `.expressions`
    "Expression",
    "ensure_expr",
    # from `functions`
    "FunctionDefinition",
    # from `.networks`
    "Network",
    "Link",
    # from `.types`
    "Type",
]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import fractions
import pathlib
import re
import tempfile

from .. import jani, model
from ..analysis import checkers


try:
    import docker
    import docker.types
except ImportError:
    raise ImportError(
        "Missing optional dependency `docker`.\n"
        "Using Storm in a Docker container requires installing `docker`."
    )


DOCKER_IMAGE = "movesrwth/storm"


# XXX: is there a better way to do this?
_result_regex = re.compile(
    r"Model checking property \"(?P<prop_name>[^\"]+)\""
    r".*?"
    r"Result \(for initial states\): (?P<prop_value>\d+(\.\d+)?)",
    flags=re.DOTALL,
)


class Toolset:
    """Interface to Storm running in a Docker container."""

    client: t.Any
    tag: str

    def __init__(self, client: t.Optional[t.Any] = None, tag: str = "travis") -> None:
        self.client = client or docker.from_env()
        self.tag = tag

    def pull(self) -> None:
        self.client.images.pull(DOCKER_IMAGE, tag=self.tag)

    def run(self, arguments: t.Sequence[str], mounts: t.Sequence[t.Any] = ()) -> str:
        """Runs storm with the provided arguments and mounts."""
        command = ["sh", "-c", "'", "./storm"]
        command.extend(arguments)
        command.extend((";", "exit 0", "'"))
        return self.client.containers.run(
            f"{DOCKER_IMAGE}:{self.tag}",
            command=" ".join(command),
            working_dir="/opt/storm/build/bin/",
            auto_remove=True,
            mounts=list(mounts),
        ).decode("utf-8")


@d.dataclass(frozen=True, eq=False)
class StormChecker(checkers.Checker):
    """Checker implementation for Storm running in Docker."""

    toolset: Toolset
    """The toolset to use."""

    engine: str = "dd"
    """The engine to use."""

    @property
    def description(self) -> str:
        return f"Storm in Docker (engine = {self.engine})"

    def check(
        self,
        network: model.Network,
        *,
        properties: t.Optional[checkers.Properties] = None,
        property_names: t.Optional[t.Iterable[str]] = None,
    ) -> checkers.Result:
        with tempfile.TemporaryDirectory(prefix="momba") as directory_name:
            input_file = pathlib.Path(directory_name) / "input.jani"
            named_properties: t.Dict[str, model.Expression] = {}
            if properties is None and property_names is None:
                named_properties.update(
                    {
                        definition.name: definition.expression
                        for definition in network.ctx.properties.values()
                    }
                )
            if properties is not None:
                named_properties.update(properties)
            if property_names is not None:
                for name in property_names:
                    named_properties[
                        name
                    ] = network.ctx.get_property_definition_by_name(name).expression
            input_file.write_text(
                jani.dump_model(network, properties=named_properties), encoding="utf-8"
            )
            output = self.toolset.run(
                (
                    "--jani",
                    "/tmp/momba-jani/input.jani",
                    "--janiproperty",
                    "--engine",
                    self.engine,
                ),
                mounts=[
                    docker.types.Mount("/tmp/momba-jani", directory_name, type="bind")
                ],
            )
        return {
            match.group("prop_name"): fractions.Fraction(match.group("prop_value"))
            for match in _result_regex.finditer(output)
        }


toolset = Toolset()

checker_sparse = StormChecker(toolset, engine="sparse")
checker_dd = StormChecker(toolset, engine="dd")

checker = checker_sparse



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import dataclasses
import enum
import functools
import json
import warnings

from .. import model
from ..model import context, expressions, operators, properties, types, functions
from ..utils import checks
from ..metadata import version


JSON = t.Union[None, int, float, str, t.Sequence["JSON"], t.Mapping[str, "JSON"]]

_JANIMap = t.Dict[str, JSON]


class ModelFeature(enum.Enum):
    """
    An enum representing optional JANI model features.
    """

    ARRAYS = "arrays"
    """ Support for arrays. """

    DATATYPES = "datatypes"
    """
    Support for datatypes.
    """

    DERIVED_OPERATORS = "derived-operators"
    """
    Support for derived operators.
    """

    EDGE_PRIORITIES = "edge-priorities"
    """
    Support for edge priorities.
    """

    FUNCTIONS = "functions"
    """
    Support for functions.
    """

    HYPERBOLIC_FUNCTIONS = "hyperbolic-functions"
    """
    Support for hyperbolic functions.
    """

    NAMED_EXPRESSIONS = "named-expressions"
    """
    Support for named expressions.
    """

    NONDET_SELECTION = "nondet-selection"
    """
    Suport for non-deterministic selection expressions.
    """

    STATE_EXIT_REWARDS = "state-exit-rewards"
    """
    Support for state exit rewards.
    """

    TRADEOFF_PROPERTIES = "tradeoff-properties"
    """
    Support for tradeoff properties.
    """

    TRIGONOMETRIC_FUNCTIONS = "trigonometric-functions"
    """
    Support for trigonometric functions.
    """

    X_MOMBA_OPERATORS = "x-momba-operators"
    """
    Support for Momba non-standard operators.
    """

    X_MOMBA_VALUE_PASSING = "x-momba-value-passing"
    """
    Support for Momba non-standard value passing.
    """


_NamedObject = t.Union[model.Location, model.Automaton]


@dataclasses.dataclass
class JANIContext:
    allow_momba_operators: bool = False

    _name_counter: int = 0

    _actions: t.Set[str] = dataclasses.field(default_factory=set)

    _names: t.Dict[_NamedObject, str] = dataclasses.field(default_factory=dict)

    _features: t.Set[ModelFeature] = dataclasses.field(default_factory=set)

    @property
    def features(self) -> t.AbstractSet[ModelFeature]:
        return self._features

    @property
    def actions(self) -> t.AbstractSet[str]:
        return self._actions

    def get_unique_name(self) -> str:
        name = f"__momba_{self._name_counter}"
        self._name_counter += 1
        return name

    def get_name(self, obj: _NamedObject) -> str:
        if obj.name is None:
            try:
                return self._names[obj]
            except KeyError:
                name = self._names[obj] = self.get_unique_name()
                return name
        return obj.name

    def require_action(self, action: str) -> str:
        self._actions.add(action)
        return action

    def require(self, feature: ModelFeature) -> None:
        self._features.add(feature)


@functools.singledispatch
def _dump_type(typ: model.Type, ctx: JANIContext) -> JSON:
    raise NotImplementedError(f"dump has not been implemented for type {typ}")


@_dump_type.register
def _dump_type_integer(typ: types.IntegerType, ctx: JANIContext) -> JSON:
    return "int"


@_dump_type.register
def _dump_type_real(typ: types.RealType, ctx: JANIContext) -> JSON:
    return "real"


@_dump_type.register
def _dump_type_bool(typ: types.BoolType, ctx: JANIContext) -> JSON:
    return "bool"


@_dump_type.register
def _dump_type_clock(typ: types.ClockType, ctx: JANIContext) -> JSON:
    return "clock"


@_dump_type.register
def _dump_type_continuous(typ: types.ContinuousType, ctx: JANIContext) -> JSON:
    return "continuous"


@_dump_type.register
def _dump_bounded_type(typ: types.BoundedType, ctx: JANIContext) -> JSON:
    jani_type: _JANIMap = {"kind": "bounded", "base": _dump_type(typ.base, ctx)}
    if typ.upper_bound:
        jani_type["upper-bound"] = _dump_prop(typ.upper_bound, ctx)
    if typ.lower_bound:
        jani_type["lower-bound"] = _dump_prop(typ.lower_bound, ctx)
    return jani_type


@_dump_type.register
def _dump_array_type(typ: types.ArrayType, ctx: JANIContext) -> JSON:
    ctx.require(ModelFeature.ARRAYS)
    return {"kind": "array", "base": _dump_type(typ.element, ctx)}


checks.check_singledispatch(
    _dump_type, types.Type, ignore={types.StateType, types.SetType}
)


@functools.singledispatch
def _dump_prop(prop: model.Expression, ctx: JANIContext) -> JSON:
    raise NotImplementedError(f"dump has not been implemented for property {prop}")


@_dump_prop.register
def _dump_identifier(expr: expressions.Name, ctx: JANIContext) -> JSON:
    return expr.identifier


@_dump_prop.register
def _dump_boolean_constant(expr: expressions.BooleanConstant, ctx: JANIContext) -> JSON:
    return expr.boolean


@_dump_prop.register
def _dump_integer_constant(expr: expressions.IntegerConstant, ctx: JANIContext) -> JSON:
    return expr.integer


@_dump_prop.register
def _dump_real_constant(expr: expressions.RealConstant, ctx: JANIContext) -> JSON:
    if isinstance(expr.real, expressions.NamedReal):
        return {"constant": expr.real.symbol}
    if not isinstance(expr.real, float) and expr.real != float(expr.real):
        warnings.warn(
            f"imprecise conversion: JSON does not support the number type {type(expr.real)}"
        )
    return float(expr.real)


@_dump_prop.register
def _dump_conditional(expr: expressions.Conditional, ctx: JANIContext) -> JSON:
    return {
        "op": "ite",
        "if": _dump_prop(expr.condition, ctx),
        "then": _dump_prop(expr.consequence, ctx),
        "else": _dump_prop(expr.alternative, ctx),
    }


_DERIVED_OPERATORS = {
    operators.BooleanOperator.IMPLY,
    operators.ComparisonOperator.GT,
    operators.ComparisonOperator.GE,
    operators.ArithmeticBinaryOperator.MIN,
    operators.ArithmeticBinaryOperator.MAX,
    operators.ArithmeticUnaryOperator.ABS,
    operators.ArithmeticUnaryOperator.SGN,
    operators.ArithmeticUnaryOperator.TRC,
}


_Transform = t.Callable[[expressions.Expression], expressions.Expression]


def normalize_xor(expr: expressions.Expression) -> expressions.Expression:
    assert (
        isinstance(expr, expressions.Boolean)
        and expr.operator is operators.BooleanOperator.XOR
    )
    return expressions.logic_or(
        expressions.logic_and(expressions.logic_not(expr.left), expr.right),
        expressions.logic_and(expr.right, expressions.logic_not(expr.left)),
    )


def normalize_equiv(expr: expressions.Expression) -> expressions.Expression:
    assert (
        isinstance(expr, expressions.Boolean)
        and expr.operator is operators.BooleanOperator.EQUIV
    )
    return expressions.logic_and(
        expressions.logic_implies(expr.left, expr.right),
        expressions.logic_implies(expr.right, expr.left),
    )


def normalize_floor_div(expr: expressions.Expression) -> expressions.Expression:
    assert (
        isinstance(expr, expressions.ArithmeticBinary)
        and expr.operator is operators.ArithmeticBinaryOperator.FLOOR_DIV
    )
    return expressions.floor(expressions.real_div(expr.left, expr.right))


_MOMBA_OPERATORS: t.Mapping[operators.BinaryOperator, _Transform] = {
    operators.BooleanOperator.XOR: normalize_xor,
    operators.BooleanOperator.EQUIV: normalize_equiv,
    operators.ArithmeticBinaryOperator.FLOOR_DIV: normalize_floor_div,
}


@_dump_prop.register
def _dump_binary_expression(
    expr: expressions.BinaryExpression, ctx: JANIContext
) -> JSON:
    if expr.operator in _DERIVED_OPERATORS:
        ctx.require(ModelFeature.DERIVED_OPERATORS)
    if expr.operator in _MOMBA_OPERATORS:
        if ctx.allow_momba_operators:
            ctx.require(ModelFeature.X_MOMBA_OPERATORS)
        else:
            return _dump_prop(_MOMBA_OPERATORS[expr.operator](expr), ctx)
    return {
        "op": expr.operator.symbol,
        "left": _dump_prop(expr.left, ctx),
        "right": _dump_prop(expr.right, ctx),
    }


@_dump_prop.register
def _dump_unary_expression(expr: expressions.UnaryExpression, ctx: JANIContext) -> JSON:
    return {"op": expr.operator.symbol, "exp": _dump_prop(expr.operand, ctx)}


@_dump_prop.register
def _dump_derivative(expr: expressions.Derivative, ctx: JANIContext) -> JSON:
    return {"op": "der", "var": expr.identifier}


@_dump_prop.register
def _dump_array_access(expr: expressions.ArrayAccess, ctx: JANIContext) -> JSON:
    ctx.require(ModelFeature.ARRAYS)
    return {
        "op": "aa",
        "exp": _dump_prop(expr.array, ctx),
        "index": _dump_prop(expr.index, ctx),
    }


@_dump_prop.register
def _dump_array_value(expr: expressions.ArrayValue, ctx: JANIContext) -> JSON:
    ctx.require(ModelFeature.ARRAYS)
    return {
        "op": "av",
        "elements": list(_dump_prop(element, ctx) for element in expr.elements),
    }


@_dump_prop.register
def _dump_array_constructor(
    expr: expressions.ArrayConstructor, ctx: JANIContext
) -> JSON:
    ctx.require(ModelFeature.ARRAYS)
    return {
        "op": "av",
        "var": expr.variable,
        "length": _dump_prop(expr.length, ctx),
        "exp": _dump_prop(expr.expression, ctx),
    }


@_dump_prop.register
def _dump_sample(expr: expressions.Sample, ctx: JANIContext) -> JSON:
    return {
        "distribution": expr.distribution.jani_name,
        "args": [_dump_prop(argument, ctx) for argument in expr.arguments],
    }


@_dump_prop.register
def _dump_selection(expr: expressions.Selection, ctx: JANIContext) -> JSON:
    ctx.require(ModelFeature.NONDET_SELECTION)
    return {
        "op": "nondet",
        "var": expr.variable,
        "exp": _dump_prop(expr.condition, ctx),
    }


@_dump_prop.register
def _dum_call(expr: functions.CallExpression, ctx: JANIContext) -> JSON:
    ctx.require(ModelFeature.FUNCTIONS)
    return {
        "op": "call",
        "function": expr.function,
        "args": [_dump_prop(argument, ctx) for argument in expr.arguments],
    }


def _dump_prop_interval(pi: properties.Interval, ctx: JANIContext) -> JSON:
    prop_interval: _JANIMap = {}
    if pi.lower is not None:
        prop_interval["lower"] = _dump_prop(pi.lower, ctx)
    if pi.lower_exclusive is not None:
        prop_interval["lower-exclusive"] = _dump_prop(pi.lower_exclusive, ctx)
    if pi.upper is not None:
        prop_interval["upper"] = _dump_prop(pi.upper, ctx)
    if pi.upper_exclusive is not None:
        prop_interval["upper-exclusive"] = _dump_prop(pi.upper_exclusive, ctx)
    return prop_interval


def _dump_reward_bound(rb: properties.RewardBound, ctx: JANIContext) -> JSON:
    return {
        "exp": _dump_prop(rb.expression, ctx),
        "accumulate": [elem.value for elem in rb.accumulate],
        "bounds": _dump_prop_interval(rb.bounds, ctx),
    }


def _dump_reward_instant(ri: properties.RewardInstant, ctx: JANIContext) -> JSON:
    return {
        "exp": _dump_prop(ri.expression, ctx),
        "accumulate": [elem.value for elem in ri.accumulate],
        "instant": _dump_prop(ri.instant, ctx),
    }


@_dump_prop.register
def _dump_filter(prop: properties.Aggregate, ctx: JANIContext) -> JSON:
    return {
        "op": "filter",
        "fun": prop.function.symbol,
        "values": _dump_prop(prop.values, ctx),
        "states": _dump_prop(prop.predicate, ctx),
    }


@_dump_prop.register
def _dump_probability(prop: properties.Probability, ctx: JANIContext) -> JSON:
    return {
        "op": "Pmax" if prop.operator is operators.MinMax.MAX else "Pmin",
        "exp": _dump_prop(prop.formula, ctx),
    }


@_dump_prop.register
def _dump_path_formula(prop: properties.PathQuantifier, ctx: JANIContext) -> JSON:
    return {
        "op": prop.quantifier.value,
        "exp": _dump_prop(prop.formula, ctx),
    }


@_dump_prop.register
def _dump_expected(prop: properties.ExpectedReward, ctx: JANIContext) -> JSON:
    expected: _JANIMap = {
        "op": "Emax" if prop.operator is operators.MinMax.MAX else "Emin",
        "exp": _dump_prop(prop.reward, ctx),
    }
    if prop.accumulate is not None:
        expected["accumulate"] = [elem.value for elem in prop.accumulate]
    if prop.reachability is not None:
        expected["reach"] = _dump_prop(prop.reachability, ctx)
    if prop.step_instant is not None:
        expected["step-instant"] = _dump_prop(prop.step_instant, ctx)
    if prop.time_instant is not None:
        expected["time-instant"] = _dump_prop(prop.time_instant, ctx)
    if prop.reward_instants is not None:
        expected["reward-instants"] = [
            _dump_reward_instant(ri, ctx) for ri in prop.reward_instants
        ]
    return expected


@_dump_prop.register
def _dump_steady(prop: properties.SteadyState, ctx: JANIContext) -> JSON:
    steady: _JANIMap = {
        "op": "Smax" if prop.operator is operators.MinMax.MAX else "Smin",
        "exp": _dump_prop(prop.formula, ctx),
    }
    if prop.accumulate is not None:
        steady["accumulate"] = [elem.value for elem in prop.accumulate]
    return steady


@_dump_prop.register
def _dump_timed(prop: properties.BinaryPathFormula, ctx: JANIContext) -> JSON:
    timed: _JANIMap = {
        "op": prop.operator.value,
        "left": _dump_prop(prop.left, ctx),
        "right": _dump_prop(prop.right, ctx),
    }
    if prop.step_bounds is not None:
        timed["step-bounds"] = _dump_prop_interval(prop.step_bounds, ctx)
    if prop.time_bounds is not None:
        timed["time-bounds"] = _dump_prop_interval(prop.time_bounds, ctx)
    if prop.reward_bounds is not None:
        timed["reward-bounds"] = [
            _dump_reward_bound(rb, ctx) for rb in prop.reward_bounds
        ]
    return timed


@_dump_prop.register
def _dump_state_selector(prop: properties.StateSelector, ctx: JANIContext) -> JSON:
    return {"op": prop.predicate.value}


@_dump_prop.register
def _dump_unary_path_formula(
    prop: properties.UnaryPathFormula, ctx: JANIContext
) -> JSON:
    jani_prop: _JANIMap = {
        "op": prop.operator.value,
        "exp": _dump_prop(prop.formula, ctx),
    }
    if prop.step_bounds is not None:
        jani_prop["step-bounds"] = _dump_prop_interval(prop.step_bounds, ctx)
    if prop.time_bounds is not None:
        jani_prop["time-bounds"] = _dump_prop_interval(prop.time_bounds, ctx)
    if prop.reward_bounds is not None:
        jani_prop["reward-bounds"] = [
            _dump_reward_bound(rb, ctx) for rb in prop.reward_bounds
        ]
    return jani_prop


checks.check_singledispatch(_dump_prop, model.Expression)


def _dump_var_decl(decl: context.VariableDeclaration, ctx: JANIContext) -> JSON:
    jani_declaration: _JANIMap = {
        "name": decl.identifier,
        "type": _dump_type(decl.typ, ctx),
    }
    if decl.is_transient is not None:
        jani_declaration["transient"] = decl.is_transient
    if decl.initial_value is not None:
        jani_declaration["initial-value"] = _dump_prop(decl.initial_value, ctx)
    return jani_declaration


def _dump_const_decl(decl: context.ConstantDeclaration, ctx: JANIContext) -> JSON:
    jani_declaration: _JANIMap = {
        "name": decl.identifier,
        "type": _dump_type(decl.typ, ctx),
    }
    if decl.value is not None:
        jani_declaration["value"] = _dump_prop(decl.value, ctx)
    return jani_declaration


def _dump_assignment(assignment: model.Assignment, ctx: JANIContext) -> JSON:
    jani_assignment: _JANIMap = {
        "ref": _dump_prop(assignment.target, ctx),
        "value": _dump_prop(assignment.value, ctx),
    }
    if assignment.index != 0:
        jani_assignment["index"] = assignment.index
    return jani_assignment


def _dump_location(loc: model.Location, ctx: JANIContext) -> JSON:
    jani_location: _JANIMap = {
        "name": ctx.get_name(loc),
        "x-momba-anonymous": loc.name is None,
    }
    if loc.progress_invariant is not None:
        jani_location["time-progress"] = {
            "exp": _dump_prop(loc.progress_invariant, ctx)
        }
    if loc.transient_values is not None:
        jani_location["transient-values"] = [
            _dump_assignment(assignment, ctx) for assignment in loc.transient_values
        ]
    return jani_location


def _dump_destination(dst: model.Destination, ctx: JANIContext) -> JSON:
    jani_destination: _JANIMap = {"location": ctx.get_name(dst.location)}
    if dst.probability is not None:
        jani_destination["probability"] = {"exp": _dump_prop(dst.probability, ctx)}
    if dst.assignments:
        jani_destination["assignments"] = [
            _dump_assignment(assignment, ctx) for assignment in dst.assignments
        ]
    return jani_destination


def _dump_edge(edge: model.Edge, ctx: JANIContext) -> JSON:
    jani_edge: _JANIMap = {
        "location": ctx.get_name(edge.location),
        "destinations": [_dump_destination(dst, ctx) for dst in edge.destinations],
    }
    if edge.action_pattern is not None:
        jani_edge["action"] = _dump_action_pattern(edge.action_pattern, ctx)
    if edge.rate is not None:
        jani_edge["rate"] = {"exp": _dump_prop(edge.rate, ctx)}
    if edge.guard is not None:
        jani_edge["guard"] = {"exp": _dump_prop(edge.guard, ctx)}
    return jani_edge


def _dump_automaton(automaton: model.Automaton, ctx: JANIContext) -> JSON:
    jani_automaton: _JANIMap = {
        "name": ctx.get_name(automaton),
        "x-momba-anonymous": automaton.name is None,
        "variables": [
            _dump_var_decl(var_decl, ctx)
            for var_decl in automaton.scope.variable_declarations
        ],
        "locations": [_dump_location(loc, ctx) for loc in automaton.locations],
        "edges": [_dump_edge(edge, ctx) for edge in automaton.edges],
        "initial-locations": [ctx.get_name(loc) for loc in automaton.initial_locations],
    }
    if automaton.initial_restriction is not None:
        jani_automaton["restrict-initial"] = {
            "exp": _dump_prop(automaton.initial_restriction, ctx)
        }
    return jani_automaton


def _dump_action_pattern(
    pattern: t.Optional[model.ActionPattern], ctx: JANIContext
) -> JSON:
    if pattern is not None:
        assert (
            not pattern.arguments
        ), "exporting action patterns with arguments to Jani is currently not supported"
        # FIXME: add support for exporting patterns with arguments
        # if pattern.identifiers:
        #     ctx.require(ModelFeature.X_MOMBA_VALUE_PASSING)
        #     return {
        #         "name": pattern.action_type.name,
        #         "identifiers": list(pattern.identifiers),
        #     }
        # else:
        return pattern.action_type.label
    return None


def _dump_link(
    instance_vector: t.Sequence[model.Instance],
    link: model.Link,
    ctx: JANIContext,
) -> JSON:
    jani_sync: _JANIMap = {
        "synchronise": [
            _dump_action_pattern(link.vector.get(instance, None), ctx)
            for instance in instance_vector
        ],
    }
    if link.result is not None:
        jani_sync["result"] = _dump_action_pattern(link.result, ctx)
    return jani_sync


def _dump_system(network: model.Network, ctx: JANIContext) -> JSON:
    instance_vector = list(network.instances)
    jani_elements: t.List[_JANIMap] = []
    for instance in instance_vector:
        jani_instance: _JANIMap = {"automaton": ctx.get_name(instance.automaton)}
        if instance.input_enable:
            jani_instance["input-enable"] = list(
                action_typ.label for action_typ in instance.input_enable
            )
        jani_elements.append(jani_instance)
    return {
        "elements": jani_elements,
        "syncs": [_dump_link(instance_vector, link, ctx) for link in network.links],
    }


def _dump_metadata(model_ctx: model.Context) -> _JANIMap:
    jani_metadata: _JANIMap = {}
    for field in {"version", "author", "description", "doi", "url"}:
        try:
            jani_metadata[field] = model_ctx.metadata[field]
        except KeyError:
            pass
    return jani_metadata


def _dump_action_type(action_type: model.ActionType, ctx: JANIContext) -> _JANIMap:
    jani_action: _JANIMap = {"name": action_type.label}
    jani_parameters: t.List[_JANIMap] = []
    for parameter in action_type.parameters:
        ctx.require(ModelFeature.X_MOMBA_VALUE_PASSING)
        jani_parameter: _JANIMap = {"type": _dump_type(parameter.typ, ctx)}
        if parameter.comment is not None:
            jani_parameter["comment"] = parameter.comment
        jani_parameters.append(jani_parameter)
    if jani_parameters:
        jani_action["parameters"] = jani_parameters
    if action_type.comment is not None:
        jani_action["comment"] = action_type.comment
    return jani_action


def dump_structure(
    network: model.Network,
    *,
    allow_momba_operators: bool = False,
    properties: t.Optional[t.Mapping[str, model.Expression]] = None,
) -> JSON:
    ctx = JANIContext(allow_momba_operators=allow_momba_operators)
    jani_metadata: t.Dict[str, str] = {}
    if properties is None:
        properties = {
            definition.name: definition.expression
            for definition in network.ctx.properties.values()
        }
    if "name" in network.ctx.metadata:
        jani_metadata
    jani_model: _JANIMap = {
        "jani-version": 1,
        "x-generator": f"Momba (v{version})",
        "x-momba-release": version,
        "name": network.name or "A Momba Model",
        "x-momba-anonymous": network.name is None,
        "metadata": _dump_metadata(network.ctx),
        "x-momba-metadata": dict(network.ctx.metadata),
        "type": network.ctx.model_type.name.lower(),
        "variables": [
            _dump_var_decl(var_decl, ctx)
            for var_decl in network.ctx.global_scope.variable_declarations
        ],
        "constants": [
            _dump_const_decl(const_decl, ctx)
            for const_decl in network.ctx.global_scope.constant_declarations
        ],
        "actions": [
            _dump_action_type(action_type, ctx)
            for action_type in network.ctx.action_types.values()
        ],
        "automata": [
            _dump_automaton(automaton, ctx) for automaton in network.ctx.automata
        ],
        "properties": [
            {"name": name, "expression": _dump_prop(prop, ctx)}
            for name, prop in properties.items()
        ],
        "system": _dump_system(network, ctx),
        # important: has to be at the end, because we collect the features while dumping
        "features": [feature.value for feature in ctx.features],
    }
    if network.initial_restriction is not None:
        jani_model["restrict-initial"] = {
            "exp": _dump_prop(network.initial_restriction, ctx)
        }
    return jani_model


def dump_model(
    network: model.Network,
    *,
    indent: t.Optional[int] = None,
    allow_momba_operators: bool = False,
    properties: t.Optional[t.Mapping[str, model.Expression]] = None,
) -> str:
    """
    Takes a Momba automaton :class:`~momba.model.Network` and returns a
    JANI string representing the network.

    The `indent` parameter controls the indentation of the resulting
    JANI string.
    `None` means no indentation.
    Set `indent` to an integer, e.g., :code:`2`, to enable pretty formatting.

    Momba supports some non-standard operators which can either be preserved
    or desugared into standard JANI.
    This behavior is controlled via the flag `allow_momba_operators`.
    If the flag is set to `True`, then the `x-momba-operators` JANI model
    feature is enabled and the non-standard operators will be preserved
    in the output.
    Otherwise, if it is set to `False` (the default case), non-standard
    operators are desugared into standard JANI.

    The `properties` parameter allows to provide an mapping from :class:`str`
    to :class:`~momba.model.Expression`.
    The provided expressions will be put as additional properties into the
    JANI-model output.
    """
    jani_structure = dump_structure(
        network, allow_momba_operators=allow_momba_operators, properties=properties
    )
    return json.dumps(
        jani_structure,
        indent=indent,
        ensure_ascii=False,
    )



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import asyncio
import json
import fractions
import pathlib
import struct
import subprocess
import tempfile

from .. import jani, model

from ..tools import modest

from . import generic
from .abstract import Oracle
from .dump_nn import dump_nn


if t.TYPE_CHECKING:
    import torch  # type: ignore


HEADER = struct.Struct("!II")  # (num_features, num_actions)
DECISION = struct.Struct("!I")  # (action,)


@d.dataclass(frozen=True)
class ModesOptions:
    max_run_length_as_end: bool = True
    additional_options: t.Sequence[str] = ()

    def apply(self, command: t.List[str]) -> None:
        if self.max_run_length_as_end:
            command.append("--max-run-length-as-end")
        command.extend(self.additional_options)


_DEFAULT_OPTIONS = ModesOptions()


def _apply_actions_observations(
    command: t.List[str], actions: generic.Actions, observations: generic.Observations
) -> None:
    if actions is generic.Actions.EDGE_BY_LABEL:
        command.append("--select-by-label")
    else:
        assert actions is generic.Actions.EDGE_BY_INDEX
    if observations is generic.Observations.LOCAL_AND_GLOBAL:
        command.append("--observations-local-global")
    elif observations is generic.Observations.OMNISCIENT:
        command.append("--observations-omniscient")
    else:
        assert observations is generic.Observations.GLOBAL_ONLY


def _create_vector_decoder(num_features: int) -> struct.Struct:
    return struct.Struct(f"!{num_features}f")


def _create_available_decoder(num_actions: int) -> struct.Struct:
    return struct.Struct(f"!{num_actions}?")


@d.dataclass()
class OracleServer:
    oracle: Oracle

    host: str = "0.0.0.0"
    port: t.Optional[int] = None

    server: t.Optional[asyncio.AbstractServer] = None

    async def _handle_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        address = writer.get_extra_info("peername")
        print(f"New connection from {address}.")
        num_features, num_actions = HEADER.unpack(await reader.readexactly(HEADER.size))
        print(f"Features: {num_features}, Actions: {num_actions}")
        state_decoder = _create_vector_decoder(num_features)
        available_decoder = _create_available_decoder(num_actions)
        try:
            while True:
                state = state_decoder.unpack(
                    await reader.readexactly(state_decoder.size)
                )
                available = available_decoder.unpack(
                    await reader.readexactly(available_decoder.size)
                )
                decision = self.oracle(state, available)
                writer.write(DECISION.pack(decision))
                await writer.drain()
        except asyncio.IncompleteReadError:
            pass

    async def start(self) -> None:
        self.server = await asyncio.start_server(
            self._handle_connection, self.host, self.port
        )
        assert self.server.sockets is not None
        self.port = self.server.sockets[0].getsockname()[1]
        await self.server.start_serving()

    async def stop(self) -> None:
        assert self.server is not None
        self.server.close()
        await self.server.wait_closed()


_NO_PARAMETERS: t.Dict[str, t.Any] = {}


def _load_result(output_file: pathlib.Path) -> t.Mapping[str, fractions.Fraction]:
    result = json.loads(output_file.read_text(encoding="utf-8-sig"))
    return {
        dataset["property"]: fractions.Fraction(dataset["value"])
        for dataset in result["data"]
        if "data" in dataset
        for dataset in dataset["data"]
        if "property" in dataset
    }


async def _check_oracle(
    model_path: pathlib.Path,
    automaton_name: str,
    oracle: Oracle,
    output_file: pathlib.Path,
    *,
    parameters: t.Mapping[str, t.Any],
    toolset: modest.Toolset,
    options: ModesOptions,
    actions: generic.Actions,
    observations: generic.Observations,
) -> None:
    server = OracleServer(oracle)
    await server.start()
    address = f"localhost:{server.port}"
    print(f"Address: {address}")
    arguments: t.List[str] = [
        "modes",
        str(model_path),
        "-O",
        str(output_file),
        "json",
        "-R",
        "oracle",
        "--socket",
        address,
        "-A",
        automaton_name,
        "--threads",
        "1",
    ]
    options.apply(arguments)
    for param_name, param_value in parameters.items():
        arguments.extend(["-E", f"{param_name}={param_value}"])
    _apply_actions_observations(arguments, actions, observations)
    process = await asyncio.subprocess.create_subprocess_exec(
        toolset.executable, *arguments
    )
    await process.wait()


async def check_oracle_async(
    network: model.Network,
    instance: model.Instance,
    oracle: Oracle,
    *,
    parameters: t.Mapping[str, t.Any] = _NO_PARAMETERS,
    toolset: modest.Toolset = modest.toolset,
    options: ModesOptions = _DEFAULT_OPTIONS,
    actions: generic.Actions = generic.Actions.EDGE_BY_INDEX,
    observations: generic.Observations = generic.Observations.GLOBAL_ONLY,
) -> t.Mapping[str, fractions.Fraction]:
    """Checks an arbitrary Python function implementing a decsion agent."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = pathlib.Path(temp_dir)
        model_path = temp_path / "model.jani"
        model_path.write_text(jani.dump_model(network), encoding="utf-8")
        output_file = temp_path / "output.json"
        assert instance.automaton.name is not None
        await _check_oracle(
            model_path,
            instance.automaton.name,
            oracle,
            output_file,
            parameters=parameters,
            toolset=toolset,
            options=options,
            actions=actions,
            observations=observations,
        )
        return _load_result(output_file)


def check_oracle(
    network: model.Network,
    instance: model.Instance,
    oracle: Oracle,
    *,
    parameters: t.Mapping[str, t.Any] = _NO_PARAMETERS,
    toolset: modest.Toolset = modest.toolset,
    options: ModesOptions = _DEFAULT_OPTIONS,
    actions: generic.Actions = generic.Actions.EDGE_BY_INDEX,
    observations: generic.Observations = generic.Observations.GLOBAL_ONLY,
) -> t.Mapping[str, fractions.Fraction]:
    """
    Checks an arbitrary Python function.

    Arguments:
        network: A JANI automaton network.
        instance: An instance of an automaton in the provided network. The decision-making
            agent trained on the resulting environment is assumed to act by resolving the
            non-determinism in this automaton.
        orcace: The Python function to check.
        parameters: Allows defining values for parameters of the JANI model.
        toolset: Specifies the location of the Modest Toolset.
        actions: Specifies the action space for the environment.
        observations: Specifies the observation space for the environment.
    """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(
        check_oracle_async(
            network,
            instance,
            oracle,
            parameters=parameters,
            toolset=toolset,
            options=options,
            actions=actions,
            observations=observations,
        )
    )


def check_nn(
    network: model.Network,
    instance: model.Instance,
    nn: torch.nn.Module,
    *,
    parameters: t.Mapping[str, t.Any] = _NO_PARAMETERS,
    toolset: modest.Toolset = modest.toolset,
    options: ModesOptions = _DEFAULT_OPTIONS,
    actions: generic.Actions = generic.Actions.EDGE_BY_INDEX,
    observations: generic.Observations = generic.Observations.GLOBAL_ONLY,
) -> t.Mapping[str, fractions.Fraction]:
    """
    Checks a PyTorch neural network.

    Arguments:
        network: A JANI automaton network.
        instance: An instance of an automaton in the provided network. The decision-making
            agent trained on the resulting environment is assumed to act by resolving the
            non-determinism in this automaton.
        nn: The PyTorch neural network.
        parameters: Allows defining values for parameters of the JANI model.
        toolset: Specifies the location of the Modest Toolset.
        actions: Specifies the action space for the environment.
        observations: Specifies the observation space for the environment.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = pathlib.Path(temp_dir)
        model_path = temp_path / "model.jani"
        model_path.write_text(jani.dump_model(network), encoding="utf-8")
        nn_path = temp_path / "nn.json"
        nn_path.write_text(dump_nn(nn))
        output_file = temp_path / "output.json"
        command: t.List[str] = [
            str(toolset.executable),
            "modes",
            str(model_path),
            "-O",
            str(output_file),
            "json",
            "-R",
            "NN",
            "-NN",
            str(nn_path),
            "-A",
            str(instance.automaton.name),
            "--threads",
            "1",
        ]
        for param_name, param_value in parameters.items():
            command.extend(["-E", f"{param_name}={param_value}"])
        options.apply(command)
        _apply_actions_observations(command, actions, observations)
        subprocess.check_call(command)
        return _load_result(output_file)



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import pathlib


Command = t.Sequence[t.Union[str, pathlib.Path]]


class ToolError(Exception):
    command: Command

    stdout: t.Optional[bytes]
    stderr: t.Optional[bytes]

    returncode: t.Optional[int]

    def __init__(
        self,
        message: str,
        command: Command = (),
        *,
        stdout: t.Optional[bytes] = None,
        stderr: t.Optional[bytes] = None,
        returncode: t.Optional[int] = None,
    ) -> None:
        super().__init__(message)
        self.command = command
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode


class ToolTimeoutError(ToolError):
    pass



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>
#
# type: ignore

from __future__ import annotations

import typing as t

import gymnasium as gym  # type: ignore
import numpy

from gymnasium import spaces  # type: ignore
from gymnasium.error import UnsupportedMode  # type: ignore

from . import abstract


class Renderer(t.Protocol):
    def render(self, state: abstract.StateVector, mode: str) -> None:
        raise NotImplementedError()


class MombaEnv(gym.Env):  # type: ignore
    """
    Implementation of an OpenAI Gym environment.

    Arguments:
        explorer: The :class:`abstract.Explorer` to use.
        renderer: An optional renderer for the OpenAI Gym API.
    """

    explorer: abstract.Explorer

    action_space: gym.Space
    observation_space: gym.Space

    renderer: t.Optional[Renderer]

    def __init__(
        self, explorer: abstract.Explorer, *, renderer: t.Optional[Renderer] = None
    ) -> None:
        super().__init__()
        self.explorer = explorer
        self.action_space = spaces.Discrete(self.explorer.num_actions)
        self.observation_space = spaces.Box(
            low=float("-inf"), high=float("inf"), shape=(self.explorer.num_features,)
        )
        self.renderer = renderer

    @property
    def available_actions(self) -> numpy.ndarray:  # type: ignore
        return numpy.array(self.explorer.available_actions)

    @property
    def available_transitions(self) -> t.Sequence[abstract.Transition]:
        return self.explorer.available_transitions

    @property
    def is_done(self) -> bool:
        return self.explorer.has_terminated

    @property
    def state_vector(self) -> numpy.ndarray:  # type: ignore
        return numpy.array(self.explorer.state_vector)

    def fork(self) -> MombaEnv:
        """Forks the environment."""
        return MombaEnv(self.explorer.fork(), renderer=self.renderer)

    def step(self, action: int) -> t.Tuple[numpy.ndarray, float, bool, t.Any]:  # type: ignore
        """Takes a decision in response to the last observation."""
        reward = self.explorer.step(action)
        state = numpy.array(self.explorer.state_vector)
        return state, reward, self.explorer.has_terminated, {}

    def reset(self) -> numpy.ndarray:  # type: ignore
        """Resets the environment to an initial state and returns an initial observation."""
        self.explorer.reset()
        return numpy.array(self.explorer.state_vector)

    def render(self, mode: str = "human") -> None:
        """Renders the environment assuming a :code:`render` has been supplied."""
        if self.renderer is None:
            raise UnsupportedMode("`MombaGym` does not support rendering")
        else:
            self.renderer.render(self.explorer.state_vector, mode)



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import dataclasses


NumberT = t.TypeVar("NumberT")


@dataclasses.dataclass(frozen=True)
class Interval(t.Generic[NumberT]):
    infimum: NumberT
    supremum: NumberT
    infimum_included: bool = True
    supremum_included: bool = True
